# pyright: strict, reportTypeCommentUsage=false, reportMissingTypeStubs=false
from datetime import datetime
import errno
import fcntl
import json
import os
import tempfile
from enum import Enum
from hashlib import md5, sha1, sha256
from itertools import chain

from metaflow._vendor.packaging.utils import (
    parse_sdist_filename,
    parse_wheel_filename,
)
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
)
from urllib.parse import urlparse, urlunparse

from metaflow.metaflow_config import CONDA_PACKAGES_DIRNAME
from metaflow.util import get_username

from .utils import (
    _ALL_PYPI_FORMATS,
    FAKEURL_PATHCOMPONENT,
    AliasType,
    arch_id,
    channel_from_url,
    correct_splitext,
    dict_to_tstr,
    get_conda_manifest_path,
    is_alias_mutable,
    tstr_to_dict,
)

# Order should be maintained
EnvID = NamedTuple("EnvID", [("req_id", str), ("full_id", str), ("arch", str)])

VALID_IMAGE_NAME_RE = "[^a-z0-9_]"


class EnvType(Enum):
    """Type of environment"""

    CONDA_ONLY = "conda-only"
    """Environment only has conda packages
    """

    PYPI_ONLY = "pypi-only"
    """Environment only has Pypi packages
    """

    PIP_ONLY = "pip-only"  # Here for legacy reasons -- we now use pypi-only
    """Alias to PYPI_ONLY
    """

    MIXED = "mixed"
    """Environment has mixed Conda and Pypi packages
    """


class TStr:
    def __init__(self, category: str, value: str):
        if category == "pip":
            # Legacy support for "pip" packages (now called pypi)
            category = "pypi"
        self._category = category
        self._value = value

    @property
    def category(self):
        return self._category

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value: str):
        self._value = value

    def __str__(self):
        return "%s::%s" % (self._category, self._value)

    def __repr__(self):
        return str(self)

    @staticmethod
    def from_str(value: str):
        splits = value.split("::", 1)
        if len(splits) != 2:
            raise ValueError("Cannot parse a TStr from %s" % value)
        c = splits[0]
        if c == "pip":
            c = "pypi"
        return TStr(c, splits[1])


def env_type_for_deps(deps: Dict[str, List[str]]) -> EnvType:
    """
    Returns the environment type based on a set of dependencies

    Parameters
    ----------
    deps : List[TStr]
        User-requested dependencies for this environment

    Returns
    -------
    EnvType
        The environment type, either CONDA_ONLY, PYPI_ONLY or MIXED
    """
    env_type = EnvType.CONDA_ONLY
    if len(deps.get("conda", [])) <= 1:
        # This is a pypi only mode
        env_type = EnvType.PYPI_ONLY
    elif deps.get("pypi", []):
        env_type = EnvType.MIXED
    return env_type


class CachePackage:
    # Cache URL explanation:
    #  - we form the cache_url based on the source URL so that we can easily check
    #    if the file is in cache.
    #  - in some cases, we want to store multiple formats for the same package (for
    #    example .conda and .tar.bz2 for conda packages or a source .tar.gz and a .whl
    #    for pypi packages)
    #  - some URLs are "fake" URLs that don't actually correspond to something we can
    #    actually download or use like local paths for locally built PYPI packages or
    #    pointers to GIT repositories (ie: there is no one file pointed to by that URL).
    #    We still use the URL as a unique identifier but mark it as fake
    #  - we therefore form the cache url using the following components ("/" separated):
    #    - pkg_type: so pypi or conda (pip as legacy)
    #    - a special marker for fake URLs (if needed)
    #    - the netloc of the base source URL
    #    - the path in the base source URL
    #    - the filename
    #    - the hash of that file
    #    - the filename
    #
    # Concretely, say our source url is https://foo/bar/baz.conda and we want to store
    # a .tar.bz2 version of the file and the .conda version of the file, in cache we
    # would have:
    #  - conda/foo/bar/baz.conda/baz.conda/<hash>/baz.conda
    #  - conda/foo/bar/baz.conda/baz.tar.bz2/<hash>/baz.tar.bz2
    #
    # If we have a GIT repository like git+https://github.com/foo/myrepo/@123#subdirectory=bar,
    # we would have:
    #  - pypi/github.com/foo/myrepo/123/bar/mypackage.whl/<hash>/mypackage.whl
    TYPE = "invalid"

    _class_per_type = None  # type: Optional[Dict[str, Type[CachePackage]]]

    @staticmethod
    def _ensure_class_per_type():
        if CachePackage._class_per_type is None:
            CachePackage._class_per_type = {
                c.TYPE: c for c in CachePackage.__subclasses__()
            }

    @classmethod
    def make_partial_cache_url(cls, base_url: str, is_real_url: bool = True):
        # This method returns the base cache URL to use (so does not include the filename
        # onwards)
        cls._ensure_class_per_type()
        url = urlparse(base_url)

        if is_real_url or url.netloc.split("/")[0] == FAKEURL_PATHCOMPONENT:
            return os.path.join(
                cast(str, CONDA_PACKAGES_DIRNAME),
                cls.TYPE,
                url.netloc,
                url.path.lstrip("/"),
            )
        else:
            return os.path.join(
                cast(str, CONDA_PACKAGES_DIRNAME),
                cls.TYPE,
                FAKEURL_PATHCOMPONENT,
                url.netloc,
                url.path.lstrip("/"),
            )

    @classmethod
    def make_cache_url(
        cls,
        base_url: str,
        filename: str,
        file_format: str,
        file_hash: str,
        is_real_url: bool = True,
    ) -> str:
        cls._ensure_class_per_type()

        filename_with_ext = "%s%s" % (filename, file_format)
        if not file_format or file_format not in cls.allowed_formats():
            raise ValueError(
                "File format '%s' for make_cache_url should be a supported file format %s"
                % (file_format, str(cls.allowed_formats()))
            )
        base_cache_url = cls.make_partial_cache_url(base_url, is_real_url)
        return os.path.join(
            base_cache_url, filename_with_ext, file_hash, filename_with_ext
        )

    def __init__(self, url: str):
        self._url = url
        basename, filename = os.path.split(url)
        _, self._pkg_fmt = correct_splitext(filename)
        if self._pkg_fmt not in self.allowed_formats():
            raise ValueError(
                "URL '%s' does not end with a supported file format %s"
                % (url, str(self.allowed_formats()))
            )
        basename, self._hash = os.path.split(basename)

    @classmethod
    def allowed_formats(cls) -> Sequence[str]:
        raise NotImplementedError()

    @property
    def url(self) -> str:
        return self._url

    @property
    def hash(self) -> str:
        return self._hash

    @property
    def format(self) -> str:
        return self._pkg_fmt  # type: ignore

    def to_dict(self) -> Dict[str, Any]:
        return {"_type": self.TYPE, "url": self._url}

    @classmethod
    def from_dict(cls, d: Mapping[str, Any]):
        cls._ensure_class_per_type()
        assert cls._class_per_type
        t = d["_type"]
        if t == "pip":
            # Support legacy files that mention "pip" packages (now called "pypi")
            t = "pypi"
        return cls._class_per_type[t](url=d["url"])

    def __str__(self):
        return "%s#%s" % (self.url, self.hash)


class CondaCachePackage(CachePackage):
    TYPE = "conda"

    @classmethod
    def allowed_formats(cls) -> Sequence[str]:
        from .utils import CONDA_FORMATS

        return CONDA_FORMATS


class PypiCachePackage(CachePackage):
    TYPE = "pypi"

    @classmethod
    def allowed_formats(cls) -> Sequence[str]:
        return _ALL_PYPI_FORMATS


class PackageSpecification:
    """
    Details about a package to be installed as part of a ResolvedEnvironment.

    If using this class in user-code, use only the attributes specified below:

    Attributes
    ----------
    filename : str
        The filename of the package
    package_name : str
        The name of the package
    package_version : str
        The version of the package
    package_detailed_version : str
        The detailed version of the package
    pkg_hashes : Iterable[Tuple[str, str]]
        The hashes of the package (for various formats if applicable)
    environment_marker:
        Environment markers allow a dependency specification to provide a rule that describes when the
        dependency should be used.

        Examples:

        In requirements.txt:
            argparse;python_version<"2.7"
        In uv.lock, it's
            [[package]]
            name = "numpy"
            resolution-markers = [
                "python_full_version >= '3.12'",
            ]
        In pylock.toml:
            [[packages]]
            name = "numpy"
            version = "2.3.4"
            marker = "python_full_version >= '3.11'"

        https://packaging.python.org/en/latest/specifications/dependency-specifiers/

    """

    TYPE = "invalid"

    _class_per_type = None  # type: Optional[Dict[str, Type[PackageSpecification]]]

    def __init__(
        self,
        filename: str,
        url: str,
        is_real_url: bool = True,
        url_format: Optional[str] = None,
        hashes: Optional[Dict[str, str]] = None,
        cache_info: Optional[Dict[str, CachePackage]] = None,
        environment_marker: Optional[str] = None,
    ):
        # Some URLs are fake when the package is built on the fly. We use the URL
        # as a unique identifier for the package so we still have a URL but it is not
        # downloadable.
        # if "/" in filename and not filename.startswith(self.TYPE):
        #     raise ValueError(
        #         "Attempting to create a package of type %s with filename %s"
        #         % (self.TYPE, filename)
        #     )
        # if self.TYPE != "conda" and not filename.startswith("%s/" % self.TYPE):  # type: ignore
        #     filename = "/".join([self.TYPE, filename])
        self._filename = filename

        self._url = url
        if url_format is None:
            url_format = correct_splitext(self._url)[1]
            if url_format not in self.allowed_formats():
                raise ValueError(
                    "URL '%s' does not end in a known ending (%s)"
                    % (self._url, str(self.allowed_formats()))
                )
        self._url_format = url_format
        self._hashes = hashes or {}
        self._cache_info = cache_info or {}
        self._environment_marker = environment_marker

        if not is_real_url:
            # If it is not a real URL, add the FAKEURL_PATHCOMPONENT but only if not
            # already there.
            url_parse_result = urlparse(self._url)
            if not url_parse_result.netloc.startswith(FAKEURL_PATHCOMPONENT):
                self._url = urlunparse(
                    (
                        url_parse_result.scheme,
                        os.path.join(FAKEURL_PATHCOMPONENT, url_parse_result.netloc),
                        url_parse_result.path,
                        url_parse_result.params,
                        url_parse_result.query,
                        url_parse_result.fragment,
                    )
                )

        (
            self._package_name,
            self._package_version,
            self._package_detailed_version,
        ) = self._split_filename()

        # Additional information used for local book-keeping as we are updating
        # the package
        self._local_dir = None  # type: Optional[str]
        self._local_path = {}  # type: Dict[str, str]
        self._is_fetched = []  # type: List[str]
        self._dirty = False

    def clone_with_filename(self, new_filename: str) -> "PackageSpecification":
        r = self.__class__(
            new_filename,
            self._url,
            self.is_downloadable_url(),
            url_format=self._url_format,
            hashes=self._hashes,
            cache_info=self._cache_info,
        )
        r._local_dir = self._local_dir
        r._local_path = self._local_path
        r._is_fetched = self._is_fetched
        return r

    @property
    def filename(self) -> str:
        return self._filename

    @property
    def package_name(self) -> str:
        return self._package_name

    def package_name_with_channel(
        self, ignore_channels: Optional[List[str]] = None
    ) -> str:
        if ignore_channels is None:
            ignore_channels = []
        if self.TYPE != "conda":
            return self.package_name
        if self.is_downloadable_url():
            # Extract the channel information from the URL. We only extract implicit
            # channels (ie: things like comet_ml) that can be embedded in the name
            # (for example: comet_ml::comet_ml)
            channel = channel_from_url(self.url)
            if channel and channel not in ignore_channels:
                return "%s::%s" % (channel, self.package_name)
        return self.package_name

    @property
    def package_version(self) -> str:
        return self._package_version

    @property
    def package_detailed_version(self) -> str:
        return self._package_detailed_version

    @property
    def url(self) -> str:
        return self._url

    @property
    def url_format(self) -> str:
        return self._url_format

    @property
    def dirty(self) -> bool:
        return self._dirty

    @classmethod
    def cache_pkg_type(cls) -> Type[CachePackage]:
        raise NotImplementedError

    def pkg_hash(self, pkg_format: str) -> Optional[str]:
        return self._hashes.get(pkg_format)

    @property
    def pkg_hashes(self) -> Iterable[Tuple[str, str]]:
        for pkg_fmt, pkg_hash in self._hashes.items():
            yield (pkg_fmt, pkg_hash)

    def add_pkg_hash(self, pkg_format: str, pkg_hash: str):
        old_hash = self.pkg_hash(pkg_format)
        if old_hash:
            if old_hash != pkg_hash:
                raise ValueError(
                    "Attempting to add an inconsistent hash for package %s; "
                    "adding %s when already have %s"
                    % (self.filename, pkg_hash, old_hash)
                )
            return
        self._dirty = True
        self._hashes[pkg_format] = pkg_hash

    @property
    def local_dir(self) -> Optional[str]:
        # Returns the local directory found for this package
        return self._local_dir

    def local_file(self, pkg_format: str) -> Optional[str]:
        # Return the local tar-ball for this package (depending on the format)
        return self._local_path.get(pkg_format)

    @property
    def local_files(self) -> Iterable[Tuple[str, str]]:
        for pkg_fmt in self.allowed_formats():
            local_path = self._local_path.get(pkg_fmt)
            if local_path:
                yield (pkg_fmt, local_path)

    @property
    def environment_marker(self) -> Optional[str]:
        return self._environment_marker

    def is_fetched(self, pkg_format: str) -> bool:
        # Return whether the local tar-ball for this package had to be fetched from
        # either cache or web
        return pkg_format in self._is_fetched

    def is_downloadable_url(self, pkg_format: Optional[str] = None) -> bool:
        if not pkg_format:
            pkg_format = self._url_format
        return pkg_format == self._url_format and not urlparse(
            self._url
        ).netloc.startswith(FAKEURL_PATHCOMPONENT)

    def is_derived(self) -> bool:
        # If the filename component of the URL does not match the filename of this package,
        # this means we derived the package from the URL in a non obvious manner
        # (transmutations don't count here -- this would return false because both
        # formats are equivalent -- this is not necessarily the case from a source
        # tar ball and a built wheel)
        url_filename_with_ext = os.path.split(urlparse(self._url).path)[1]
        url_filename = correct_splitext(url_filename_with_ext)[0]
        return url_filename != self._filename

    def can_add_filename(self, filename_with_ext: str) -> bool:
        # Tests if a filename is a compatible filename for this package. This is used
        # when there are multiple possibilities with PYPI packages for example
        raise NotImplementedError

    def add_local_dir(self, local_path: str):
        # Add a local directory that is present for this package
        local_dir = self.local_dir
        if local_dir:
            if local_dir != local_path:
                raise ValueError(
                    "Attempting to add an inconsistent local directory for package %s; "
                    "adding %s when already have %s"
                    % (self.filename, local_path, local_dir)
                )
            return
        self._dirty = True
        self._local_dir = local_path

    def add_local_file(
        self,
        pkg_format: str,
        local_path: str,
        pkg_hash: Optional[str] = None,
        downloaded: bool = False,
        replace: bool = False,
    ):
        # Add a local file for this package indicating whether it was downloaded
        existing_path = self.local_file(pkg_format)
        if not replace and existing_path and local_path != existing_path:
            raise ValueError(
                "Attempting to add inconsistent local files of format %s for a package %s; "
                "adding %s when already have %s"
                % (pkg_format, self.filename, local_path, existing_path)
            )

        known_hash = self._hashes.get(pkg_format)
        added_hash = pkg_hash or self.hash_pkg(local_path)
        if not replace and known_hash and known_hash != added_hash:
            raise ValueError(
                "Attempting to add inconsistent local files of format %s for package %s; "
                "got a hash of %s but expected %s"
                % (pkg_format, self.filename, added_hash, known_hash)
            )
        self._dirty = replace or existing_path is None or known_hash is None
        self._local_path[pkg_format] = local_path
        self._hashes[pkg_format] = added_hash

        if downloaded and pkg_format not in self._is_fetched:
            self._dirty = True
            self._is_fetched.append(pkg_format)

    def cached_version(self, pkg_format: str) -> Optional[CachePackage]:
        return self._cache_info.get(
            self._url_format if pkg_format == "_any" else pkg_format
        )

    @property
    def cached_versions(self) -> Iterable[Tuple[str, CachePackage]]:
        for pkg_fmt, cached in self._cache_info.items():
            yield (pkg_fmt, cached)

    def add_cached_version(self, pkg_format: str, cache_info: CachePackage) -> None:
        if cache_info.TYPE != self.TYPE:
            raise ValueError(
                "Attempting to add a cache package of type %s for a package of type %s"
                % (cache_info.TYPE, self.TYPE)
            )

        old_cache_info = self.cached_version(pkg_format)
        if old_cache_info:
            if (
                old_cache_info.url != cache_info.url
                or old_cache_info.hash != cache_info.hash
            ):
                raise ValueError(
                    "Attempting to add inconsistent cache information for format %s of package %s; "
                    "adding %s when already have %s"
                    % (pkg_format, self.filename, cache_info, old_cache_info)
                )
        else:
            self._dirty = True
            self._cache_info[pkg_format] = cache_info

        old_pkg_hash = self.pkg_hash(pkg_format)
        if old_pkg_hash:
            if old_pkg_hash != cache_info.hash:
                raise ValueError(
                    "Attempting to add inconsistent cache information for format %s of package %s; "
                    "adding a package with hash %s when expected %s"
                    % (pkg_format, self.filename, cache_info.hash, old_pkg_hash)
                )
        else:
            self._dirty = True
            self._hashes[pkg_format] = cache_info.hash

    def is_cached(self, formats: List[str]) -> bool:
        if formats:
            return all(
                [
                    self._url_format if f == "_any" else f in self._cache_info
                    for f in formats
                ]
            )
        return len(self._cache_info) > 0

    def to_dict(self) -> Dict[str, Any]:
        d = {
            "_type": self.TYPE,
            "filename": self.filename,
            "url": self.url,
            "url_format": self.url_format,
            "hashes": self._hashes,
        }  # type: Dict[str, Any]
        if self._cache_info:
            cache_d = {
                pkg_format: info.to_dict()
                for pkg_format, info in self._cache_info.items()
            }
            d["cache_info"] = cache_d
        return d

    def _split_filename(self) -> Tuple[str, str, str]:
        raise NotImplementedError

    @classmethod
    def from_dict(cls, d: Mapping[str, Any]):
        if cls._class_per_type is None:
            cls._class_per_type = {c.TYPE: c for c in cls.__subclasses__()}

        cache_info = d.get("cache_info", {})  # type: Dict[str, Any]
        url_format = d["url_format"]
        t = d["_type"]
        if t == "pip":
            # Legacy support for "pip" packages, now called "pypi"
            t = "pypi"
        return cls._class_per_type[t](
            filename=d["filename"],
            url=d["url"],
            url_format=url_format,
            hashes=d["hashes"],
            cache_info={
                pkg_fmt: CachePackage.from_dict(info)
                for pkg_fmt, info in cache_info.items()
            },
        )

    @classmethod
    def allowed_formats(cls) -> Sequence[str]:
        raise NotImplementedError()

    @classmethod
    def base_hash(cls):
        raise NotImplementedError()

    @classmethod
    def base_hash_name(cls) -> str:
        raise NotImplementedError()

    @classmethod
    def hash_pkg(cls, path: str) -> str:
        base_hash = cls.base_hash()
        with open(path, "rb") as f:
            for byte_block in iter(lambda: f.read(8192), b""):
                base_hash.update(byte_block)
        return base_hash.hexdigest()

    def __str__(self):
        return "<%s package: %s>" % (self.TYPE, self.filename)


class CondaPackageSpecification(PackageSpecification):
    TYPE = "conda"

    @classmethod
    def cache_pkg_type(cls):
        return CondaCachePackage

    @classmethod
    def allowed_formats(cls) -> Sequence[str]:
        from .utils import CONDA_FORMATS

        return CONDA_FORMATS

    @classmethod
    def base_hash(cls):
        return md5()

    @classmethod
    def base_hash_name(cls) -> str:
        return "md5"

    def can_add_filename(self, filename_with_ext: str) -> bool:
        # Tests if a filename is a compatible filename for this package. This is used
        # when there are multiple possibilities with PYPI packages for example
        ext = correct_splitext(filename_with_ext)[1]
        return ext in self.allowed_formats()

    def _split_filename(self) -> Tuple[str, str, str]:
        pkg, v, addl = self._filename.rsplit("-", 2)
        return pkg, v, "-".join([v, addl])


class PypiPackageSpecification(PackageSpecification):
    TYPE = "pypi"

    @classmethod
    def cache_pkg_type(cls):
        return PypiCachePackage

    @classmethod
    def allowed_formats(cls) -> Sequence[str]:
        return _ALL_PYPI_FORMATS

    @classmethod
    def base_hash(cls):
        return sha256()

    @classmethod
    def base_hash_name(cls) -> str:
        return "sha256"

    def can_add_filename(self, filename_with_ext: str) -> bool:
        # Tests if a filename is a compatible filename for this package. This is used
        # when there are multiple possibilities with PYPI packages for example
        base_filename, ext = correct_splitext(filename_with_ext)

        if ext == ".whl":
            # This will make sure the wheel matches the tags and what not
            return base_filename == self.filename
        else:
            # Source packages are always ok to add. It may have a different name
            return ext in _ALL_PYPI_FORMATS

    def _split_filename(self) -> Tuple[str, str, str]:
        try:
            # Try a source distribution first. We don't know which to try because the
            # url format may now repersent what is actually in this package (the URL
            # may be to the source but this represents a wheel).
            name, version = parse_sdist_filename("".join([self._filename, ".tar.gz"]))
            return (str(name), str(version), "")
        except ValueError:
            name, version, buildtag, _ = parse_wheel_filename(
                "".join([self._filename, ".whl"])
            )
            return (
                str(name),
                str(version),
                "" if len(buildtag) == 0 else "%s-%s" % (buildtag[0], buildtag[1]),
            )

    def is_external_url(self, sources: Dict[str, List[str]]) -> bool:
        # Returns True if the URL is not in the sources
        # This is used to determine if we should consider a package as a local package to
        # be downloaded before resolving the environment
        pypi_sources = sources.get("pypi", [])
        return not any(urlparse(source).netloc in self.url for source in pypi_sources)


class ResolvedEnvironment:
    """
    A resolved environment fully describes a Conda/Pypi/Mixed environment including
    the initial dependencies that led to this environment as well as all the transitive
    dependencies.
    """

    def __init__(
        self,
        user_dependencies: Dict[str, List[str]],
        user_sources: Optional[Dict[str, List[str]]],
        user_extra_args: Optional[Dict[str, List[str]]],
        arch: Optional[str] = None,
        env_id: Optional[EnvID] = None,
        all_packages: Optional[Sequence[PackageSpecification]] = None,
        resolved_on: Optional[datetime] = None,
        resolved_by: Optional[str] = None,
        co_resolved: Optional[List[str]] = None,
        env_type: EnvType = EnvType.MIXED,
        accurate_source: bool = True,
        full_id_unique_keys: Optional[Dict[str, str]] = None,
        # Sample format: {'uv_lock_content': '33746974cb80adfea6047b5c5713c59c0dbc721da3179501c1bad261241cd3d3'}
    ):
        self._env_type = env_type
        self._user_dependencies = dict_to_tstr(user_dependencies)
        # We sort the user dependencies as the order does not matter and it makes
        # it easier for users to find their dependencies
        self._user_dependencies.sort(key=lambda k: k.value)
        self._user_sources = dict_to_tstr(user_sources) if user_sources else []
        self._user_extra_args = dict_to_tstr(user_extra_args) if user_extra_args else []

        self._accurate_source = accurate_source
        self._full_id_unique_keys = full_id_unique_keys or {}

        if not env_id:
            env_req_id = ResolvedEnvironment.get_req_id(
                user_dependencies,
                user_sources or {},
                user_extra_args or {},
            )

            env_full_id = "_unresolved"
            if all_packages is not None:
                env_full_id = self._compute_full_id(
                    all_packages, arch or arch_id(), self._full_id_unique_keys
                )

            self._env_id = EnvID(
                req_id=env_req_id, full_id=env_full_id, arch=arch or arch_id()
            )
        else:
            self._env_id = env_id
        self._all_packages = list(all_packages) if all_packages else []
        self._resolved_on = resolved_on or datetime.now()
        self._resolved_by = (
            resolved_by or cast(Optional[str], get_username()) or "unknown"
        )
        self._co_resolved = co_resolved or [arch or arch_id()]
        self._parent = None  # type: Optional["CachedEnvironmentInfo"]
        self._dirty = False

    @staticmethod
    def get_req_id(
        deps: Dict[str, List[str]],
        sources: Optional[Dict[str, List[str]]] = None,
        extra_args: Optional[Dict[str, List[str]]] = None,
    ) -> str:
        if sources is None:
            sources = {}
        if extra_args is None:
            extra_args = {}
        return ResolvedEnvironment._compute_hash(
            chain(
                (str(TStr(c, pkg)) for c in sorted(deps) for pkg in sorted(deps[c])),
                (
                    str(TStr(c, pkg))
                    for c in sorted(sources)
                    for pkg in sorted(sources[c])
                ),
                (
                    str(TStr(c, arg))
                    for c in sorted(extra_args)
                    for arg in sorted(extra_args[c])
                ),
            )
        )

    @staticmethod
    def set_coresolved_full_id(envs: Sequence["ResolvedEnvironment"]) -> None:
        envs = sorted(envs, key=lambda x: x.env_id.arch)
        to_hash = []  # type: List[str]
        archs = []  # type: List[str]
        for env in envs:
            archs.append(env.env_id.arch)
            to_hash.append(env.env_id.arch)
            to_hash.extend(
                [
                    "%s#%s" % (p.filename, p.pkg_hash(p.url_format))
                    for p in sorted(env.packages, key=lambda p: p.filename)
                ]
            )
        new_full_id = ResolvedEnvironment._compute_hash(to_hash)
        for env in envs:
            env.set_coresolved(archs, new_full_id)

    @property
    def is_info_accurate(self) -> bool:
        # Returns True if the requirements for this environment are accurate. This is
        # currently always the case. This code was initially added in case we could
        # get an environment from a FULL_ID which doesn't guarantee uniqueness of the
        # requirements part (different requirements can resolve to the same environment)
        return self._accurate_source

    @property
    def deps(self) -> List[TStr]:
        """
        Returns the user-requested dependencies for this environment.

        Returns
        -------
        List[TStr]
            List of user requested dependencies. Each dependency will indicate if it
            is from the pypi ecosystem or conda one.
        """
        return self._user_dependencies

    @property
    def sources(self) -> List[TStr]:
        """
        Returns the user-requested sources to use for this environment. In the
        Conda ecosystem, these are the channels and in the Pypi ecosystem, these
        are repository URLs.

        Returns
        -------
        List[TStr]
            List of user requested sources.
        """
        return self._user_sources

    @property
    def extras(self) -> List[TStr]:
        """
        List any user-requested "extras". This is currently only applicable for the Pypi
        ecosystem.

        Returns
        -------
        List[TStr]
            List of user requested extras.
        """
        return self._user_extra_args

    @property
    def env_id(self) -> EnvID:
        """
        Returns the full environment ID for this environment.

        Returns
        -------
        EnvID
            Unique identifier for this environment.
        """
        if self._env_id.full_id in ("_default", "_unresolved") and self._all_packages:
            env_full_id = self._compute_full_id(
                self._all_packages,
                self._env_id.arch or arch_id(),
                self._full_id_unique_keys,
            )

            self._env_id = self._env_id._replace(full_id=env_full_id)
        return self._env_id

    @property
    def packages(self) -> Iterable[PackageSpecification]:
        """
        Returns the full list of packages that will be installed for this environment.
        The order of the packages returned will be the order in which they will
        be installed.

        Yields
        ------
        Iterator[PackageSpecification]
            Packages that will be installed.
        """
        # This returns the packages in the order in which they were added which
        # corresponds to the installation order. In some cases, the installation order
        # matters so we make sure to preserve it.
        for p in self._all_packages:
            yield p

    @property
    def resolved_on(self) -> datetime:
        """
        Returns the data and time on which this environment was resolved.

        Returns
        -------
        datetime
            Date/time on which this environment was resolved.
        """
        return self._resolved_on

    @property
    def resolved_by(self) -> str:
        """
        Returns the user who triggered the resolution of this environment.

        Returns
        -------
        str
            Username who triggered the resolution of this environment.
        """
        return self._resolved_by

    @property
    def co_resolved_archs(self) -> List[str]:
        """
        Lists all the architectures for which this environment was resolved.

        Note that there is *no* guarantee that the environments are identical. They
        are *likely* identical as they were resolved at the same time.

        Returns
        -------
        List[str]
            Architectures for which this environment was simultaneously resolved.
        """
        return self._co_resolved

    @property
    def env_type(self) -> EnvType:
        """
        Type of the environment. One of "conda", "pypi" or "mixed".

        Returns
        -------
        EnvType
            Environment type.
        """
        return self._env_type

    @property
    def dirty(self) -> bool:
        if self._dirty:
            return True
        return any((p.dirty for p in self._all_packages))

    @dirty.setter
    def dirty(self, value: bool):
        self._dirty = value

    def set_parent(self, parent: "CachedEnvironmentInfo"):
        self._parent = parent

    def add_package(self, pkg: PackageSpecification):
        self._dirty = True
        self._all_packages.append(pkg)
        if self._env_id.full_id not in ("_default", "_unresolved"):
            self._env_id._replace(full_id="_unresolved")

    def set_coresolved(self, archs: List[str], full_id: str) -> None:
        self._dirty = True
        self._env_id = EnvID(
            req_id=self._env_id.req_id, full_id=full_id, arch=self._env_id.arch
        )
        self._co_resolved = archs

    def pretty_print(self, local_instances: Optional[List[str]]) -> str:
        lines = []  # type: List[str]
        pypi_packages = []  # type: List[PackageSpecification]
        conda_packages = []  # type: List[PackageSpecification]
        for p in self.packages:
            if p.TYPE == "pypi":
                pypi_packages.append(p)
            else:
                conda_packages.append(p)

        pypi_packages.sort(key=lambda x: x.package_name)
        conda_packages.sort(key=lambda x: x.package_name)

        if self._parent:
            immutable_aliases, mutable_aliases = self._parent.aliases_for_env(
                self.env_id
            )
            immutable_aliases = [
                a for a in immutable_aliases if a != self.env_id.full_id
            ]
            is_default = (
                self._parent.get_default(self.env_id.req_id, self.env_id.arch)
                == self.env_id
            )
        else:
            immutable_aliases = []
            mutable_aliases = []
            is_default = False

        lines.append(
            "*%sEnvironment of type %s full hash* %s:%s"
            % (
                "DEFAULT " if is_default else "",
                self.env_type.value,
                self.env_id.req_id,
                self.env_id.full_id,
            )
        )
        if immutable_aliases or mutable_aliases:
            lines.append(
                "*Aliases* %s"
                % ", ".join(
                    chain(
                        map(lambda x: ":".join(x.rsplit("/", 1)), immutable_aliases),
                        (
                            "%s (mutable)" % ":".join(a.rsplit("/", 1))
                            for a in mutable_aliases
                        ),
                    )
                )
            )
        lines.extend(
            [
                "*Arch* %s" % self.env_id.arch,
                "*Available on* %s" % ", ".join(self.co_resolved_archs),
                "",
                "*Resolved on* %s" % self.resolved_on,
                "*Resolved by* %s" % self.resolved_by,
                "",
            ]
        )
        if local_instances:
            lines.extend(["*Locally present as* %s" % ", ".join(local_instances), ""])

        # The `replace` is because the echo function uses "*" to split in bold/not bold
        # It is quite likely to have a package spec like "3.8.*" so we replace it with
        # the similar looking "3.8.x"
        lines.append(
            "*User-requested packages* %s"
            % ", ".join([str(d).replace("*", "x") for d in self.deps])
        )

        if self.sources:
            lines.append(
                "*User sources* %s" % ", ".join([str(s) for s in self.sources])
            )

        if self.extras:
            lines.append(
                "*Extra resolution flags* %s" % ", ".join([str(s) for s in self.extras])
            )
        lines.append("")

        if conda_packages:
            lines.append(
                "*Conda Packages installed* %s"
                % ", ".join(
                    [
                        "%s==%s"
                        % (
                            p.package_name,
                            p.package_detailed_version,
                        )
                        for p in conda_packages
                    ]
                )
            )

        if pypi_packages:
            lines.append(
                "*Pypi Packages installed* %s"
                % ", ".join(
                    [
                        "%s==%s" % (p.package_name, p.package_version)
                        for p in pypi_packages
                    ]
                )
            )
        lines.append("")
        return "\n".join(lines)

    def quiet_print(self, local_instances: Optional[List[str]] = None) -> str:
        pypi_packages = []  # type: List[PackageSpecification]
        conda_packages = []  # type: List[PackageSpecification]
        for p in self.packages:
            if p.TYPE == "pypi":
                pypi_packages.append(p)
            else:
                conda_packages.append(p)

        pypi_packages.sort(key=lambda x: x.package_name)
        conda_packages.sort(key=lambda x: x.package_name)

        if self._parent:
            immutable_aliases, mutable_aliases = self._parent.aliases_for_env(
                self.env_id
            )
            immutable_aliases = [
                a for a in immutable_aliases if a != self.env_id.full_id
            ]
        else:
            immutable_aliases = []
            mutable_aliases = []

        aliases = ",".join(
            chain(
                map(lambda x: ":".join(x.rsplit("/", 1)), immutable_aliases),
                ("%s(m)" % ":".join(a.rsplit("/", 1)) for a in mutable_aliases),
            )
        )
        return "%s %s %s %s %s %s %s %s %s conda:%s pypi:%s %s" % (
            self.env_id.req_id,
            self.env_id.full_id,
            aliases,
            self.env_id.arch,
            self.resolved_on.isoformat(),
            self.resolved_by,
            ",".join(self.co_resolved_archs),
            ",".join([str(d) for d in self.deps]),
            ",".join([str(s) for s in self.sources]) if self.sources else "NONE",
            ",".join(
                [
                    "%s==%s" % (p.package_name, p.package_detailed_version)
                    for p in conda_packages
                ]
            ),
            ",".join(
                ["%s==%s" % (p.package_name, p.package_version) for p in pypi_packages]
            ),
            ",".join(local_instances) if local_instances else "NONE",
        )

    def is_cached(self, formats: Dict[str, List[str]]) -> bool:
        return all([pkg.is_cached(formats.get(pkg.TYPE, [])) for pkg in self.packages])

    def to_dict(self) -> Dict[str, Any]:
        """
        Returns a dictionary representation of this environment.

        Returns
        -------
        Dict[str, Any]
            Dictionary representation of this environment. Note that the environent
            ID is not stored so needs to be stored separately if `from_dict` needs
            to be used later.
        """
        return {
            "deps": [str(x) for x in self._user_dependencies],
            "sources": [str(x) for x in self._user_sources],
            "extras": [str(x) for x in self._user_extra_args],
            "packages": [p.to_dict() for p in self.packages],
            "resolved_on": self._resolved_on.isoformat(),
            "resolved_by": self._resolved_by,
            "resolved_archs": self._co_resolved,
            "env_type": self._env_type.value,
            "accurate_source": self._accurate_source,
            "full_id_unique_keys": self._full_id_unique_keys,
        }

    @classmethod
    def from_dict(
        cls,
        env_id: EnvID,
        d: Mapping[str, Any],
    ) -> "ResolvedEnvironment":
        """
        Creates a ResolvedEnvironment from a representation dumped by `to_dict` and
        an environment ID.

        Parameters
        ----------
        env_id : EnvID
            Environment ID for this environment.
        d : Mapping[str, Any]
            Representation for this environment (dumped with `to_dict`).

        Returns
        -------
        ResolvedEnvironment
            A resolved environment.
        """
        all_packages = [PackageSpecification.from_dict(pd) for pd in d["packages"]]
        env_type = EnvType(d.get("env_type", EnvType.MIXED.value))
        # Backward compatible aliasing
        if env_type == EnvType.PIP_ONLY:
            env_type = EnvType.PYPI_ONLY
        return cls(
            user_dependencies=tstr_to_dict([TStr.from_str(x) for x in d["deps"]]),
            user_sources=tstr_to_dict([TStr.from_str(x) for x in d["sources"]]),
            user_extra_args=tstr_to_dict(
                [TStr.from_str(x) for x in d.get("extras", [])]
            ),
            env_id=env_id,
            all_packages=all_packages,
            resolved_on=datetime.fromisoformat(d["resolved_on"]),
            resolved_by=d["resolved_by"],
            co_resolved=d["resolved_archs"],
            env_type=env_type,
            accurate_source=d.get("accurate_source", True),
            full_id_unique_keys=d.get("full_id_unique_keys", None),
        )

    @staticmethod
    def _compute_hash(inputs: Iterable[str]) -> str:
        return sha1(b" ".join([s.encode("ascii") for s in inputs])).hexdigest()

    @staticmethod
    def _unique_keys_to_list(
        unique_keys: Dict[str, str],
    ) -> List[str]:
        # Using "{k}:{v}" not "{k}#{v}" to avoid mixing with package filename#hash,
        # in case this helps debuggability.
        return [f"{k}:{v}" for k, v in sorted(unique_keys.items())]

    # There are two levels of hashing for the environment:
    # * the req_id which is based on the user-requested dependencies and sources.
    # * the full_id which is based on the resolved user dependencies.
    #
    # Originally, full_id is composed of only the resolved dependencies and the
    # architecture, and a req_id can be mapped to a recently resolved full_id.
    # This justifies the caching mechanism such as the "_default" key under the
    # req_id entry in the environment manifest.
    #
    # However, as we introduced uv,
    # we found operations such as `uv lock --upgrade` can introduce different
    # resolved dependencies while keeping the same user-requested dependencies.
    # As a result, full_id needs to take into consideration full_id_unique_keys
    # as well.
    @staticmethod
    def _compute_full_id(
        all_packages: Sequence[PackageSpecification],
        arch: str,
        full_id_unique_keys: Optional[Dict[str, str]] = None,
    ) -> str:
        to_hash_list = [
            "%s#%s" % (p.filename, p.pkg_hash(p.url_format))
            for p in sorted(all_packages, key=lambda p: p.filename)
        ] + [arch]
        to_hash_list.extend(
            ResolvedEnvironment._unique_keys_to_list(full_id_unique_keys)
            if full_id_unique_keys
            else []
        )

        env_full_id = ResolvedEnvironment._compute_hash(to_hash_list)
        return env_full_id


class MetaflowResolvedEnvironment:
    """
    Resolved Metaflow environment.
    """

    def __init__(self, env: ResolvedEnvironment):
        self._resolved_env = env

    @property
    def deps(self) -> List[TStr]:
        """
        Returns the user-requested dependencies for this environment.

        Returns
        -------
        List[TStr]
            List of user requested dependencies. Each dependency will indicate if it
            is from the pypi ecosystem or conda one.
        """
        return self._resolved_env.deps

    @property
    def sources(self) -> List[TStr]:
        """
        Returns the user-requested sources to use for this environment. In the
        Conda ecosystem, these are the channels and in the Pypi ecosystem, these
        are repository URLs.

        Returns
        -------
        List[TStr]
            List of user requested sources.
        """
        return self._resolved_env.sources

    @property
    def extras(self) -> List[TStr]:
        """
        List any user-requested "extras". This is currently only applicable for the Pypi
        ecosystem.

        Returns
        -------
        List[TStr]
            List of user requested extras.
        """
        return self._resolved_env.extras

    @property
    def env_id(self) -> EnvID:
        """
        Returns the full environment ID for this environment.

        Returns
        -------
        EnvID
            Unique identifier for this environment.
        """
        return self._resolved_env.env_id

    @property
    def packages(self) -> Iterable[PackageSpecification]:
        """
        Returns the full list of packages that will be installed for this environment.
        The order of the packages returned will be the order in which they will
        be installed.

        Yields
        ------
        Iterator[PackageSpecification]
            Packages that will be installed.
        """
        for p in self._resolved_env.packages:
            yield p

    @property
    def resolved_on(self) -> datetime:
        """
        Returns the data and time on which this environment was resolved.

        Returns
        -------
        datetime
            Date/time on which this environment was resolved.
        """
        return self._resolved_env.resolved_on

    @property
    def resolved_by(self) -> str:
        """
        Returns the user who triggered the resolution of this environment.

        Returns
        -------
        str
            Username who triggered the resolution of this environment.
        """
        return self._resolved_env.resolved_by

    @property
    def co_resolved_archs(self) -> List[str]:
        """
        Lists all the architectures for which this environment was resolved.

        Note that there is *no* guarantee that the environments are identical. They
        are *likely* identical as they were resolved at the same time.

        Returns
        -------
        List[str]
            Architectures for which this environment was simultaneously resolved.
        """
        return self._resolved_env.co_resolved_archs

    @property
    def env_type(self) -> EnvType:
        """
        Type of the environment. One of "conda", "pypi" or "mixed".

        Returns
        -------
        EnvType
            Environment type.
        """
        return self._resolved_env.env_type

    def to_dict(self) -> Dict[str, Any]:
        """
        Returns a dictionary representation of this environment.

        Returns
        -------
        Dict[str, Any]
            Dictionary representation of this environment. Note that the environent
            ID is not stored so needs to be stored separately if `from_dict` needs
            to be used later.
        """
        return self._resolved_env.to_dict()


class CachedEnvironmentInfo:
    def __init__(
        self,
        version: int = 1,
        step_mappings: Optional[Dict[str, Tuple[Optional[str], ...]]] = None,
        env_mutable_aliases: Optional[Dict[str, Tuple[str, ...]]] = None,
        env_aliases: Optional[Dict[str, Tuple[str, ...]]] = None,
        resolved_environments: Optional[
            Dict[str, Dict[str, Dict[str, Union[ResolvedEnvironment, str]]]]
        ] = None,
    ):
        self._version = version
        self._step_mappings: Dict[str, Tuple[Optional[str], ...]] = (
            step_mappings if step_mappings else {}
        )
        self._env_mutable_aliases = env_mutable_aliases if env_mutable_aliases else {}
        self._env_aliases = env_aliases if env_aliases else {}
        self._resolved_environments = (
            resolved_environments if resolved_environments else {}
        )

    def set_default(self, env_id: EnvID):
        per_arch_envs = self._resolved_environments.get(env_id.arch)
        if per_arch_envs is None:
            raise ValueError("No cached environments for %s" % env_id.arch)
        per_req_id_envs = per_arch_envs.get(env_id.req_id)
        if per_req_id_envs is None:
            raise ValueError(
                "No cached environments for requirement ID: %s" % env_id.req_id
            )
        per_req_id_envs["_default"] = env_id.full_id

    def get_default(self, req_id: str, arch: Optional[str]) -> Optional[EnvID]:
        arch = arch or arch_id()
        per_arch_envs = self._resolved_environments.get(arch)
        if per_arch_envs is None:
            return None
        per_req_id_envs = per_arch_envs.get(req_id)
        if per_req_id_envs is None:
            return None
        full_id = per_req_id_envs.get("_default")
        if full_id is None:
            return None
        elif isinstance(full_id, str):
            return EnvID(req_id=req_id, full_id=full_id, arch=arch)
        return full_id.env_id

    def clear_default(self, req_id: str, arch: Optional[str]) -> None:
        arch = arch or arch_id()
        per_arch_envs = self._resolved_environments.get(arch)
        if per_arch_envs is None:
            return
        per_req_id_envs = per_arch_envs.get(req_id)
        if per_req_id_envs is None:
            return
        if "_default" in per_req_id_envs:
            del per_req_id_envs["_default"]

    def add_resolved_env(self, env: ResolvedEnvironment):
        env_id = env.env_id
        per_arch_envs = self._resolved_environments.setdefault(env_id.arch, {})
        per_req_id_envs = per_arch_envs.setdefault(env_id.req_id, {})
        per_req_id_envs[env_id.full_id] = env

        env.set_parent(self)
        # Update _env_aliases adding the full_id so we can look up an environment
        # directly by the full_id
        v = self._env_aliases.setdefault(
            env_id.full_id, (env_id.req_id, env_id.full_id)
        )
        if v != (env_id.req_id, env_id.full_id):
            # This can happen if two req_ids resolve to the same full_id
            # For now we ignore and we will just use the first resolved environment
            # for that full-id
            pass

    def add_alias(
        self, alias_type: AliasType, resolved_alias: str, req_id: str, full_id: str
    ):
        if alias_type == AliasType.PATHSPEC:
            v = self._step_mappings.setdefault(resolved_alias, (req_id, full_id))
            if v != (req_id, full_id):
                raise ValueError(
                    "Assigning (%s, %s) to step '%s' but already have (%s, %s)"
                    % (req_id, full_id, resolved_alias, v[0], v[1])
                )
        elif alias_type == AliasType.FULL_ID or alias_type == AliasType.GENERIC:
            if is_alias_mutable(alias_type, resolved_alias):
                self._env_mutable_aliases[resolved_alias] = (req_id, full_id)
            else:
                # At this point, this is not a mutable alias
                v = self._env_aliases.setdefault(resolved_alias, (req_id, full_id))
                if v != (req_id, full_id):
                    raise ValueError("Alias '%s' is not mutable" % resolved_alias)
        # Missing AliasType.REQ_FULL_ID but we don't record aliases for that.

    def env_for(
        self,
        req_id: str,
        full_id: str = "_default",
        arch: Optional[str] = None,
        full_id_unique_keys: Optional[Dict[str, str]] = None,
    ) -> Optional[ResolvedEnvironment]:
        arch = arch or arch_id()
        per_arch_envs = self._resolved_environments.get(arch)
        if per_arch_envs:
            per_req_id_envs = per_arch_envs.get(req_id)
            if per_req_id_envs:
                if full_id == "_default":
                    full_id = per_req_id_envs.get("_default", "_invalid")  # type: ignore
                env = per_req_id_envs.get(full_id)  # type: ignore

                if (
                    env
                    and isinstance(env, ResolvedEnvironment)
                    and env._full_id_unique_keys != full_id_unique_keys
                ):
                    env = None
                return env  # type: ignore

        return None

    def envs_for(
        self, req_id: str, arch: Optional[str] = None
    ) -> Iterator[Tuple[EnvID, ResolvedEnvironment]]:
        arch = arch or arch_id()
        per_arch_envs = self._resolved_environments.get(arch)
        if per_arch_envs:
            per_req_id_envs = per_arch_envs.get(req_id)
            if per_req_id_envs:
                for env in per_req_id_envs.values():
                    if isinstance(env, ResolvedEnvironment):
                        yield env.env_id, env
                    # The other case is that env is a string (when the key in
                    # per_req_id_envs is "_default" but we will list that
                    # when we get to the actual environment for it so we don't output
                    # anything.

    def env_id_for_alias(
        self, alias_type: AliasType, resolved_alias: str, arch: Optional[str] = None
    ) -> Optional[EnvID]:
        req_id: Optional[str]
        full_id: Optional[str]
        if alias_type == AliasType.REQ_FULL_ID:
            req_id, full_id = resolved_alias.rsplit(":", 1)
        elif alias_type == AliasType.PATHSPEC:
            req_id, full_id = self._step_mappings.get(resolved_alias, (None, None))
        elif alias_type == AliasType.GENERIC or alias_type == AliasType.FULL_ID:
            req_id, full_id = self._env_aliases.get(
                resolved_alias,
                self._env_mutable_aliases.get(resolved_alias, (None, None)),
            )
        else:
            raise ValueError("Alias type not supported")
        if req_id and full_id:
            return EnvID(req_id=req_id, full_id=full_id, arch=arch or arch_id())
        return None

    def env_for_alias(
        self,
        alias_type: AliasType,
        resolved_alias: str,
        arch: Optional[str] = None,
    ) -> Optional[ResolvedEnvironment]:
        env_id = self.env_id_for_alias(alias_type, resolved_alias, arch)
        if env_id:
            return self.env_for(resolved_alias[0], resolved_alias[1], arch)
        return None

    def aliases_for_env(self, env_id: EnvID) -> Tuple[List[str], List[str]]:
        # Returns immutable and mutable aliases
        # We don't cache this as this is fairly infrequent use (just CLI)
        immutable_aliases = [
            k
            for k, v in self._env_aliases.items()
            if v == (env_id.req_id, env_id.full_id)
        ]
        mutable_aliases = [
            k
            for k, v in self._env_mutable_aliases.items()
            if v == (env_id.req_id, env_id.full_id)
        ]
        return (immutable_aliases, mutable_aliases)

    @property
    def envs(self) -> Iterator[Tuple[EnvID, ResolvedEnvironment]]:
        for arch, per_arch_envs in self._resolved_environments.items():
            for req_id, per_req_id_envs in per_arch_envs.items():
                for full_id, env in per_req_id_envs.items():
                    if isinstance(env, ResolvedEnvironment):
                        yield (
                            EnvID(req_id=req_id, full_id=full_id, arch=arch),
                            env,
                        )

    def update(self, cached_info: "CachedEnvironmentInfo"):
        if self._version != cached_info._version:
            raise ValueError(
                "Cannot update an environment of version %d with one of version %d"
                % (self._version, cached_info._version)
            )
        self._step_mappings.update(cached_info._step_mappings)
        self._env_mutable_aliases.update(cached_info._env_mutable_aliases)
        self._env_aliases.update(cached_info._env_aliases)

        for arch, per_arch_envs in cached_info._resolved_environments.items():
            per_arch_resolved_env = self._resolved_environments.setdefault(arch, {})
            for req_id, per_req_id_envs in per_arch_envs.items():
                per_req_id_resolved_env = per_arch_resolved_env.setdefault(req_id, {})
                per_req_id_resolved_env.update(per_req_id_envs)

    def to_dict(self) -> Dict[str, Any]:
        result = {
            "version": 1,
            "mappings": self._step_mappings,
            "mutable_aliases": self._env_mutable_aliases,
            "aliases": self._env_aliases,
        }  # type: Dict[str, Any]
        resolved_envs: Dict[str, Dict[str, Dict[str, Any]]] = {}
        for arch, per_arch_envs in self._resolved_environments.items():
            per_arch_resolved_env = {}
            for req_id, per_req_id_envs in per_arch_envs.items():
                per_req_id_resolved_env: Dict[str, Any] = {}
                for full_id, env in per_req_id_envs.items():
                    if isinstance(env, ResolvedEnvironment):
                        per_req_id_resolved_env[full_id] = env.to_dict()
                    elif full_id == "_unresolved":
                        # This is a non resolved environment -- this should not happen
                        continue
                    else:
                        # This is for the special "_default" key
                        per_req_id_resolved_env[full_id] = env
                per_arch_resolved_env[req_id] = per_req_id_resolved_env
            resolved_envs[arch] = per_arch_resolved_env
        result["environments"] = resolved_envs
        return result

    @classmethod
    def from_dict(cls, d: Mapping[str, Any]):
        version = int(d.get("version", -1))
        if version != 1:
            raise ValueError("Wrong version information for CachedInformationInfo")

        step_mappings = d.get(
            "mappings", {}
        )  # type: Dict[str, Tuple[Optional[str], ...]]
        aliases = d.get("aliases", {})  # type: Dict[str, Tuple[str, ...]]
        mutable_aliases = d.get(
            "mutable_aliases", {}
        )  # type: Dict[str, Tuple[str, ...]]

        for k, v in step_mappings.items():
            step_mappings[k] = tuple(v)
        for k, v in aliases.items():
            aliases[k] = tuple(v)
        for k, v in mutable_aliases.items():
            mutable_aliases[k] = tuple(v)

        resolved_environments_dict = d.get("environments", {})
        resolved_environments = (
            {}
        )  # type: Dict[str, Dict[str, Dict[str, Union[ResolvedEnvironment, str]]]]
        for arch, per_arch_envs in resolved_environments_dict.items():
            # Parse the aliases first to make sure we have all of them
            resolved_per_req_id = {}
            for req_id, per_req_id_envs in per_arch_envs.items():
                resolved_per_full_id = (
                    {}
                )  # type: Dict[str, Union[ResolvedEnvironment, str]]
                for full_id, env in per_req_id_envs.items():
                    if full_id == "_default":
                        # Special key meaning to use this fully resolved environment
                        # for anything that matches env_id but does not have a specific
                        # mapping
                        resolved_per_full_id[full_id] = env
                    else:
                        resolved_env = ResolvedEnvironment.from_dict(
                            EnvID(req_id=req_id, full_id=full_id, arch=arch), env
                        )
                        resolved_per_full_id[full_id] = resolved_env
                resolved_per_req_id[req_id] = resolved_per_full_id
            resolved_environments[arch] = resolved_per_req_id
        me = cls(
            version,
            step_mappings=step_mappings,
            env_mutable_aliases=mutable_aliases,
            env_aliases=aliases,
            resolved_environments=resolved_environments,
        )
        for per_arch in resolved_environments.values():
            for per_req in per_arch.values():
                for env in per_req.values():
                    if isinstance(env, ResolvedEnvironment):
                        env.set_parent(me)

        return me


def read_conda_manifest(ds_root: str) -> CachedEnvironmentInfo:
    path = get_conda_manifest_path(ds_root)
    if os.path.exists(path) and os.path.getsize(path) > 0:
        with os.fdopen(os.open(path, os.O_RDONLY), "r", encoding="utf-8") as f:
            try:
                fcntl.flock(f, fcntl.LOCK_SH)
                return CachedEnvironmentInfo.from_dict(json.load(f))
            finally:
                fcntl.flock(f, fcntl.LOCK_UN)
    else:
        return CachedEnvironmentInfo()


def write_to_conda_manifest(ds_root: str, info: CachedEnvironmentInfo):
    path = get_conda_manifest_path(ds_root)
    try:
        os.makedirs(os.path.dirname(path))
    except OSError as x:
        if x.errno != errno.EEXIST:
            raise

    # Create temporary file in same directory for atomic rename
    temp_f = None
    try:
        with tempfile.NamedTemporaryFile(
            mode="w", encoding="utf-8", dir=os.path.dirname(path), delete=False
        ) as temp_f:
            with os.fdopen(
                os.open(path, os.O_RDWR | os.O_CREAT), "r+", encoding="utf-8"
            ) as f:
                try:
                    fcntl.flock(f, fcntl.LOCK_EX)
                    # We first read the file and then update it with the info we have
                    # This makes sure that if the file was updated by another process between
                    # the time we read it and now, we have a sum of all information.
                    # This can happen if multiple runs are running on the same machine
                    # from the same directory.
                    current_content = CachedEnvironmentInfo()
                    if os.path.getsize(path) > 0:
                        # Not a new file
                        f.seek(0)
                        current_content = CachedEnvironmentInfo.from_dict(json.load(f))

                    # Merge with new info
                    current_content.update(info)

                    # Write to temporary file first so we can do an atomic swap
                    # of the file
                    json.dump(current_content.to_dict(), temp_f)
                    temp_f.flush()
                    os.fsync(temp_f.fileno())

                    os.rename(temp_f.name, path)
                finally:
                    fcntl.flock(f, fcntl.LOCK_UN)
    finally:
        # Clean up temp file if it still exists (e.g., if rename failed)
        if temp_f and os.path.exists(temp_f.name):
            try:
                os.unlink(temp_f.name)
            except OSError:
                pass



# pyright: strict, reportTypeCommentUsage=false, reportMissingTypeStubs=false
from __future__ import annotations

import email.parser
import email.policy
import hashlib
import os
import platform
import re
import shutil
import subprocess
import tempfile
from enum import Enum
from itertools import chain
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    FrozenSet,
    Iterable,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
)
from urllib.parse import unquote, urlparse

import metaflow.metaflow_config as mf_config
from metaflow._vendor.packaging.markers import Marker, default_environment
from metaflow._vendor.packaging.requirements import Requirement
from metaflow._vendor.packaging.specifiers import SpecifierSet
from metaflow._vendor.packaging.tags import (
    Tag,
    _cpython_abis,
    compatible_tags,
    cpython_tags,
    mac_platforms,
)
from metaflow._vendor.packaging.utils import BuildTag, parse_wheel_filename
from metaflow._vendor.packaging.version import Version, InvalidVersion
from metaflow.debug import debug
from metaflow.exception import MetaflowException
from metaflow.metaflow_config import CONDA_BUILDER_ENV_PACKAGES  # type: ignore
from metaflow.metaflow_config import CONDA_MAGIC_FILE_V2  # type: ignore
from metaflow.metaflow_config import CONDA_PREFERRED_FORMAT  # type: ignore
from metaflow.metaflow_config import CONDA_SRCS_AUTH_INFO  # type: ignore
from metaflow.metaflow_config import CONDA_SYS_DEFAULT_GPU_PACKAGES  # type: ignore
from metaflow.metaflow_config import CONDA_SYS_DEFAULT_PACKAGES  # type: ignore
from metaflow.metaflow_config import CONDA_SYS_DEPENDENCIES  # type: ignore
from metaflow.metaflow_config import CONDA_SYS_MARKERS  # type: ignore
from metaflow.metaflow_environment import InvalidEnvironmentException
from metaflow.util import walk_without_cycles

from requests import PreparedRequest, Response, Session
from requests.auth import AuthBase, HTTPBasicAuth

if TYPE_CHECKING:
    import metaflow_extensions.nflx.plugins.conda.env_descr
    from metaflow_extensions.netflix_ext.plugins.conda.env_descr import (
        PackageSpecification,
        ResolvedEnvironment,
    )

# NOTA: Most of the code does not assume that there are only two formats BUT the
# transmute code does (since you can only specify the infile -- the outformat and file
# are inferred)
_ALL_CONDA_FORMATS: Tuple[str, ...] = (".tar.bz2", ".conda")
# NOTE: Order is important as it is a preference order
_ALL_PYPI_FORMATS = (".whl", ".tar.gz", ".zip")
_VALID_IMAGE_NAME = "[^-a-z0-9_/]"
_VALID_TAG_NAME = "[^-a-z0-9_]"

# Things that need to be put in any of the builder environments
_BUILDER_ENVS_PACKAGES = (
    "pip",
    "wheel",
    "tomli",
    "setuptools",
    "uv<=0.7.8; python_version<'3.8'",
    "uv ; python_version>='3.8'",
)


class AliasType(Enum):
    PATHSPEC = "pathspec"
    FULL_ID = "full-id"
    REQ_FULL_ID = "both-id"
    GENERIC = "generic"


# List of formats that guarantees the preferred format is first. This is important as
# functions that rely on selecting the "preferred" source of a package rely on the
# preferred format being first.
if CONDA_PREFERRED_FORMAT and CONDA_PREFERRED_FORMAT != "none":
    CONDA_FORMATS = (
        CONDA_PREFERRED_FORMAT,
        *[x for x in _ALL_CONDA_FORMATS if x != CONDA_PREFERRED_FORMAT],
    )  # type: Tuple[str, ...]
else:
    CONDA_FORMATS = _ALL_CONDA_FORMATS  # noqa: F811
FAKEURL_PATHCOMPONENT = "_fake"

_double_equal_match = re.compile(r"(==\*(?=\s|;|,|$))|(==(?=[<=>!~]))")
_build_marker_match = re.compile(r"([^;]*?)([-=][a-zA-Z0-9_]+)(?=\s|;|,|$)")


class CondaException(MetaflowException):
    headline = "Conda ran into an error while setting up environment."

    def __init__(self, error: Union[Sequence[Exception], str]):
        if isinstance(error, list):
            error = "\n".join([str(x) for x in error])
        super(CondaException, self).__init__(error)  # type: ignore


class CondaStepException(CondaException):
    def __init__(self, exception: CondaException, steps: Sequence[str]):
        msg = "Step(s): {steps}, Error: {error}".format(
            steps=steps, error=exception.message
        )
        super(CondaStepException, self).__init__(msg)


def get_builder_envs_dep(orig_deps: List[str]) -> List[str]:
    builder_packages = {}  # type: Dict[str, Dict[Optional["Marker"], SpecifierSet]]
    for pkg_req in chain(_BUILDER_ENVS_PACKAGES, CONDA_BUILDER_ENV_PACKAGES, orig_deps):
        r = get_requirement(pkg_req)
        builder_packages.setdefault(r.name, {})[r.marker] = (
            builder_packages.setdefault(r.name, {}).setdefault(r.marker, SpecifierSet())
            & r.specifier
        )
    to_return: List[str] = []
    for pkg_name, pkg_markers in builder_packages.items():
        for marker, version_spec in pkg_markers.items():
            marker_str = ";" + str(marker) if marker else ""
            if version_spec == SpecifierSet():
                to_return.append(f"{pkg_name}{marker_str}")
            else:
                to_return.append(
                    f"{pkg_name}=={str(version_spec).lstrip('=')}{marker_str}"
                )
    return to_return


def convert_filepath(path: str, file_format: Optional[str] = None) -> Tuple[str, str]:
    if file_format and not path.endswith(file_format):
        path, cur_ext = correct_splitext(path)
        if cur_ext not in _ALL_CONDA_FORMATS:
            raise ValueError(
                "URL '%s' does not end with a supported file format %s"
                % (path, str(_ALL_CONDA_FORMATS))
            )
        path = "%s%s" % (path, file_format)
    return os.path.split(path)


def get_conda_manifest_path(ds_root: str) -> str:
    return os.path.join(ds_root, CONDA_MAGIC_FILE_V2)  # type: ignore


def get_conda_root(datastore_type: str) -> str:
    conda_root = getattr(mf_config, "CONDA_%sROOT" % datastore_type.upper())
    if conda_root is None:
        # We error on METAFLOW_DATASTORE_SYSROOT_<ds> because that is the default used
        raise MetaflowException(
            msg="METAFLOW_DATASTORE_SYSROOT_%s must be set!" % datastore_type.upper()
        )
    debug.conda_exec("Conda root is at %s" % conda_root)
    return conda_root  # type: ignore


def arch_id() -> str:
    bit = "32"
    if platform.machine().endswith("64"):
        bit = "64"
    if platform.system() == "Linux":
        return "linux-%s" % bit
    elif platform.system() == "Darwin":
        # Support M1 Mac
        if platform.machine() == "arm64":
            return "osx-arm64"
        else:
            return "osx-%s" % bit
    else:
        raise InvalidEnvironmentException(
            "The *@conda* decorator is not supported "
            "outside of Linux and Darwin platforms"
        )


_global_python_versions: Optional[List[str]] = None


# Dynamically generate acceptable Python versions as major.minor.patch format.


def _generate_python_versions() -> List[str]:
    min_major: int = 3
    max_major: int = 3
    min_minor: int = 7
    max_minor: int = 15
    min_patch: int = 0
    max_patch: int = 25

    global _global_python_versions
    if _global_python_versions is not None:
        return _global_python_versions
    versions: List[str] = []
    for major in range(min_major, max_major + 1):
        minor_range = range(min_minor, max_minor + 1)
        for minor in minor_range:
            for patch in range(min_patch, max_patch + 1):
                versions.append(f"{major}.{minor}.{patch}")
    _global_python_versions = versions
    return versions


def get_sys_packages(
    virtual_packages: Dict[str, str],
    arch_requested: str,
    gpu_requested: bool,
) -> Dict[str, str]:
    # For now we handle __cuda and __glibc.

    # The goal of this function is to take the list of virtual_packages and to return
    # the sys packages we need to add to the user dependencies (so packages that start with
    # sys::). The reason for having them as part of user packages is that we can then
    # use them in computing the user request hash (since these will impact the packages
    # that are installed). The sys:: packages will then be properly added when resolving
    # the environment as virtual packages (either through CONDA_OVERRIDE_XXX or directly
    # as part of a virtual_packages.yml for conda-lock)
    if virtual_packages and arch_requested == arch_id():
        result = dict(virtual_packages)
    else:
        result = dict(CONDA_SYS_DEFAULT_PACKAGES.get(arch_requested, {}))

    if gpu_requested:
        result.update(CONDA_SYS_DEFAULT_GPU_PACKAGES)

    # Clean up
    result = {
        k: v[:-2] if v.endswith("=0=0") else (v + "=0" if not v.endswith("=0") else v)
        for k, v in result.items()
        if k in CONDA_SYS_DEPENDENCIES
    }
    return result


def get_glibc_version() -> Optional[str]:
    # Uses ldd which prints the glibc version on the system and has the version
    # as the last element in the first line.
    try:
        output = subprocess.check_output("ldd --version", shell=True)
        vers_line = output.splitlines()[0]
        version = vers_line.split(b" ")[-1]
        return version.decode("utf-8")
    except Exception:
        return None


# TODO: refactor pip_resolver and conda_lock_resolver into using this method.
def get_maximum_glibc_version(architecture: str, deps: Dict[str, List[str]]) -> str:
    # The upstream source for the architecture is arch_id(), which only returns
    # "linux-64", "linux-32", "osx-arm64", or "osx-64". Note that "linux-aarch64" is mapped
    # to "linux-64" by arch_id().
    #
    # The downstream function pypi_tags_from_arch() further refines the architecture and may
    # handle cases like "linux-aarch64". For example outputs of pypi_tags_from_arch(), see:
    # https://github.netflix.net/gist/shuishiy/af1f3287731931b443fddecb005fb7a6 (Linux)
    # https://github.netflix.net/gist/shuishiy/e78e7004369cca43a14f1efd266bdc0a (macOS)
    if architecture != "linux-64":
        return ""

    glibc_version_list = [d for d in deps.get("sys", []) if d.startswith("__glibc")]

    if len(glibc_version_list) != 1:
        raise CondaException("Could not determine maximum GLIBC version")
    # Get version looking like 2.27=0
    glibc_version = glibc_version_list[0][len("__glibc==") :]
    # Strip =0
    glibc_version = glibc_version.split("=", 1)[0]
    glibc_version = glibc_version.replace(".", "_")
    return glibc_version


def get_python_full_version_from_builder_envs(
    builder_envs: Optional[List["ResolvedEnvironment"]] = None,
    architecture: Optional[str] = None,
) -> Optional[str]:
    if not architecture:
        raise ValueError(
            "Architecture must be specified in get_python_full_version_from_builder_envs."
        )

    if not builder_envs:
        raise ValueError(
            "Builder environments must be specified in get_python_full_version_from_builder_envs."
        )

    builder_env = [r for r in builder_envs if r.env_id.arch == architecture][0]

    python_full_version = None
    for p in builder_env.packages:
        if p.filename.startswith("python-"):
            python_full_version = p.package_version
            break

    # Verify this is the full version.
    # Since the package name should be from Conda, it should always be a full version.
    # This gist https://github.netflix.net/gist/e9bd9c1624ad9de8c3dfbd956c2fd295 is
    # the result of "conda search python".
    if python_full_version and python_full_version.count(".") != 2:
        raise ValueError(
            f"Python version {python_full_version} is not a full version (missing patch version)"
        )

    return python_full_version


def pypi_tags_from_arch(
    python_version: str, arch: str, glibc_version: str
) -> List[Tag]:
    # Converts a Conda architecture to a tuple containing (implementation, platforms, abis)
    # This function will assume a CPython implementation

    # This is inspired by what pip does:
    # https://github.com/pypa/pip/blob/0442875a68f19b0118b0b88c747bdaf6b24853ba/src/pip/_internal/utils/compatibility_tags.py
    py_version = tuple(map(int, python_version.split(".")[:2]))
    if arch == "linux-64":
        max_glibc = "_%s" % glibc_version
        platforms = []  # type: List[str]
        for s in (
            "1",
            "2010",
            "2014",
            "_2_17",
            "_2_18",
            "_2_19",
            "_2_20",
            "_2_21",
            "_2_23",
            "_2_24",
            "_2_25",
            "_2_26",
            "_2_27",
            "_2_28",
            "_2_29",
            "_2_30",
            "_2_31",
            "_2_32",
            "_2_33",
            "_2_34",
            "_2_35",
            "_2_36",
            "_2_37",
            "_2_38",
        ):
            platforms.append("manylinux%s_x86_64" % s)
            if s == max_glibc:
                break

        platforms.append("linux_x86_64")
    elif arch == "osx-64":
        platforms = mac_platforms((12, 0), "x86_64")
    elif arch == "osx-arm64":
        platforms = mac_platforms((12, 0), "arm64")
    else:
        raise InvalidEnvironmentException("Unsupported platform: %s" % arch)

    interpreter = "cp%s" % ("".join(map(str, py_version)))

    abis = _cpython_abis(py_version)

    supported = []  # type: List[Tag]
    supported.extend(cpython_tags(py_version, abis, platforms))
    supported.extend(compatible_tags(py_version, interpreter, platforms))

    return supported


ParseExplicitResult = NamedTuple(
    "ParseExplicitResult",
    [("filename", str), ("url", str), ("url_format", str), ("hash", Optional[str])],
)


def parse_explicit_url_conda(url: str) -> ParseExplicitResult:
    # Takes a URL in the form url#hash and returns:
    #  - the filename
    #  - the URL (without the hash)
    #  - the format for the URL
    #  - the hash
    url_clean, url_hash = url.rsplit("#", 1)
    filename, url_format = correct_splitext(os.path.split(urlparse(url_clean).path)[1])
    if url_format not in _ALL_CONDA_FORMATS:
        raise CondaException(
            "URL '%s' is not a supported format (%s)" % (url, CONDA_FORMATS)
        )
    filename = unquote(filename)
    return ParseExplicitResult(
        filename=filename, url=url_clean, url_format=url_format, hash=url_hash
    )


def parse_explicit_path_pypi(
    path: str, hash_value: Optional[str] = None
) -> ParseExplicitResult:
    # Prevent circular import
    from .env_descr import PypiPackageSpecification

    # Takes a filename in the form file://<path> and returns:
    #  - the filename
    #  - the URL (always file://local-<hash>/<filename> so there is no way another
    #    build/user conflicts. We consider them to be all distinct.
    #  - the format of the URL
    #  - the hash will be set to None
    if not path.startswith("file://"):
        raise CondaException("Local path '%s' does not start with file://" % path)
    path = path[7:]
    orig_filename, url_format = correct_splitext(os.path.basename(path))
    if url_format not in _ALL_PYPI_FORMATS:
        raise CondaException(
            "Path '%s' is not a supported format (%s)" % (path, str(_ALL_PYPI_FORMATS))
        )
    if hash_value is None:
        if not os.path.isfile(path):
            # This is not a file so we can't get a hash. Instead, we will compute
            # the last modified time for files within the directory containing it. This only happens
            # when the user has a local directory and is trying to build a package
            # from there. The modified time will act a bit like a hash to determine
            # if the package being built has changed. If it has, the hash will be different
            # and the "fake URL" will be different so we won't find the package
            # in the cache. If it hasn't changed, the hash will be the same and we will
            # avoid randomly rebuilding the package.
            hash_value = hash_directory(os.path.dirname(path))
        else:
            hash_value = PypiPackageSpecification.hash_pkg(path)
    return ParseExplicitResult(
        filename=unquote(orig_filename),
        url="file://local-%s/%s" % (hash_value, orig_filename),
        url_format=url_format,
        hash=hash_value,
    )


def parse_explicit_url_pypi(url: str) -> ParseExplicitResult:
    # Takes a URL in the form url#hash or url and returns:
    #  - the filename
    #  - the URL (without the hash)
    #  - the format for the URL
    #  - the hash
    # NOTE: The URL may not have a hash in some cases in which case we need to account
    # for the fact that the rsplit below produces only ONE element.
    parts = url.rsplit("#", 1)
    url_clean = parts[0]
    url_hash = parts[1] if len(parts) == 2 else None
    if url_hash:
        if not url_hash.startswith("sha256="):
            raise CondaException("URL '%s' has a SHA type which is not supported" % url)
        url_hash = url_hash[7:]
    else:
        # Standardize to None in the off-chance that the url is <url># without anything
        # after the #.
        url_hash = None

    # TODO: refactor: call parse_explicit_url_pypi_no_hash() instead.
    filename, url_format = correct_splitext(os.path.split(urlparse(url_clean).path)[1])
    if url_format not in _ALL_PYPI_FORMATS:
        raise CondaException(
            "URL '%s' is not a supported format (%s)" % (url, str(_ALL_PYPI_FORMATS))
        )
    filename = unquote(filename)
    return ParseExplicitResult(
        filename=filename, url=url_clean, url_format=url_format, hash=url_hash
    )


def parse_explicit_url_pypi_no_hash(url: str) -> ParseExplicitResult:
    filename, url_format = correct_splitext(os.path.split(urlparse(url).path)[1])

    if url_format not in _ALL_PYPI_FORMATS:
        raise CondaException(
            "URL '%s' is not a supported format (%s)" % (url, str(_ALL_PYPI_FORMATS))
        )

    filename = unquote(filename)
    return ParseExplicitResult(
        filename=filename,
        url=url,
        url_format=url_format,
        hash=None,
    )


def plural_marker(count: int) -> str:
    return "s" if count != 1 else ""


def is_hexadecimal(s: str) -> bool:
    return not re.search("[^0-9a-f]", s)


def resolve_env_alias(env_alias: str) -> Tuple[AliasType, str]:
    if env_alias.startswith("step:"):
        # Pathspec environment alias
        env_alias = env_alias[5:]
        env_alias_splits = env_alias.split("/")
        if len(env_alias_splits) in (3, 4):
            return AliasType.PATHSPEC, "/".join(env_alias_splits[:3])
    elif len(env_alias) == 81 and env_alias[40] == ":":
        # Full environment hashes are:
        #  - a 40 character requirement hash
        #  - a colon
        #  - a 40 character full environment hash
        # thus a total size of 81 characters with the 40th character being a ":"
        req_id, full_id = env_alias.split(":", 1)
        if is_hexadecimal(req_id) and is_hexadecimal(full_id):
            return AliasType.REQ_FULL_ID, env_alias
    elif len(env_alias) == 40 and is_hexadecimal(env_alias):
        # Just a full ID environment hash
        # For now we do not support this -- remove if you want to support
        # The issue with supporting this is that a full-id can refer to multiple req-id
        # so it is impossible to unambiguously identify the source environment (ie:
        # the user requested dependencies, etc)
        raise MetaflowException(
            "Invalid format for environment alias: '%s'" % env_alias
        )
        # return AliasType.FULL_ID, env_alias
    else:
        # An alias in the form <name>:<tag> (where <tag> is optional and assumed to
        # be "latest" if not present)
        splits = env_alias.rsplit(":", 1)
        if len(splits) == 2:
            image_name = splits[0]
            image_tag = splits[1]
        else:
            image_name = env_alias
            image_tag = "latest"
        if re.search(_VALID_IMAGE_NAME, image_name):
            raise MetaflowException(
                "An environment name must contain only "
                "lowercase alphanumeric characters, dashes, underscores and forward slashes."
            )
        if image_name[0] == "/" or image_name[-1] == "/":
            raise MetaflowException(
                "An environment name must not start or end with '/'"
            )
        if re.search(_VALID_TAG_NAME, image_tag):
            raise MetaflowException(
                "An environment tag name must contain only "
                "lowercase alphanumeric characters, dashes and underscores."
            )
        return AliasType.GENERIC, "/".join([image_name, image_tag])
    raise MetaflowException("Invalid format for environment alias: '%s'" % env_alias)


def is_alias_mutable(alias_type: AliasType, resolved_alias: str) -> bool:
    if alias_type != AliasType.GENERIC:
        return False
    splits = resolved_alias.rsplit("/", 1)
    return len(splits) == 2 and splits[1] in ("latest", "candidate", "stable")


def dict_to_tstr(
    deps: Dict[str, List[str]],
) -> List["metaflow_extensions.nflx.plugins.conda.env_descr.TStr"]:
    from .env_descr import TStr  # Avoid circular import

    result = []  # type: List[TStr]
    for category, values in deps.items():
        result.extend([TStr(category, v) for v in values])
    return result


def tstr_to_dict(
    deps: List["metaflow_extensions.nflx.plugins.conda.env_descr.TStr"],
) -> Dict[str, List[str]]:
    result = {}  # type: Dict[str, List[str]]
    for dep in deps:
        result.setdefault(dep.category, []).append(dep.value)
    return result


def split_into_dict(deps: List[str]) -> Dict[str, str]:
    result = {}  # type: Dict[str, str]
    for dep in deps:
        s = dep.split("==", 1)
        if len(s) == 1 or s[1] == "*":
            result[s[0]] = ""
        else:
            result[s[0]] = s[1]
    return result


def dict_to_strlist(deps: Dict[str, str]) -> List[str]:
    to_return: List[str] = []
    for dep_name, dep_specifier in deps.items():
        dep_specifier = dep_specifier.lstrip()
        if dep_specifier:
            if dep_specifier[0] == ";":
                # Marker only
                to_return.append(f"{dep_name}{dep_specifier}")
            else:
                to_return.append(f"{dep_name}=={dep_specifier}")
        else:
            to_return.append(dep_name)
    return to_return


def get_requirement(dep_str: str) -> Requirement:
    # Conda packages contain extra information in the version_string and we can
    # also have double equals and channel information. We need to remove all this to
    # get a valid requirement string.
    # For example:
    # - channel::package==version -> package==version
    # - package==<=version => package<=version
    # - package==version=foo -> package==version
    dep_str = dep_str.split("::", 1)[-1]  # Remove channel information
    dep_str = _double_equal_match.sub("", dep_str)  # Remove double equals
    # Note that if no match, the sub is a no-op which is what we want.
    dep_str = _build_marker_match.sub(r"\1", dep_str)  # Remove build marker
    return Requirement(dep_str)


def clean_up_double_equal(deps: Iterable[str]) -> List[str]:
    return [_double_equal_match.sub("", d) for d in deps]


def merge_dep_dicts(
    d1: Dict[str, str], d2: Dict[str, str], only_last_deps: bool = False
) -> Dict[str, str]:
    # Merge dictionaries of version constraints taking *all* constraints
    # into account instead of replacing them.
    # Order will be preserved based on the order of the first dict
    # only_last_deps will only return things that are already in d2. This is to merge with
    # anything in d1 but only return things that are in d2. This is used to compute
    # the new set of dependencies introduced by the user.
    result = {}  # type: Dict[str, str]
    for key in set(d2.keys() if only_last_deps else chain(d1.keys(), d2.keys())):
        v1 = [v for v in d1.get(key, "").split(",") if v and v != "*"]
        v2 = [v for v in d2.get(key, "").split(",") if v and v != "*"]
        sv1 = set([v.lstrip("=") for v in v1])
        # We dedup so that if we have the exact same constraints, we don't change
        # things. We try to keep the order of the first dict
        v2 = [v for v in v2 if v.lstrip("=") not in sv1]

        # We need to make sure that things that don't start with ~, <, >, ! or = get
        # == if they are not first.
        resulting_versions = list(chain(v1, v2))
        if resulting_versions:
            resulting_versions = [resulting_versions[0]] + [
                v if v[0] in ("~", "<", ">", "!", "=") else "==%s" % v
                for v in resulting_versions[1:]
            ]
        result[key] = ",".join(resulting_versions)
    return result


def reform_pypi_filename(
    name: str, version: Version, build: BuildTag, tags: FrozenSet[Tag]
) -> str:
    if build:
        build_str = "%d%s" % (build[0], build[1])
    else:
        build_str = None
    # We need to iterate over all the tags to form the string
    # interp1.interp2-abi1.abi2-platform1.platform2
    interpreters, abis, platforms = zip(
        *map(lambda x: (x.interpreter, x.abi, x.platform), tags)
    )
    tag_str = "-".join(
        [
            ".".join(sorted(interpreters)),
            ".".join(sorted(abis)),
            ".".join(sorted(platforms)),
        ]
    )

    if build_str:
        return "-".join([str(name), str(version), build_str, tag_str])
    return "-".join([str(name), str(version), tag_str])


def correct_splitext(filename_with_ext: str) -> Tuple[str, str]:
    # This handles .tar.gz and .tar.bz2 which is what we care about.
    # Tried with pathlib too but that has other issues since . is a valid character
    # in conda packages
    all_splits = filename_with_ext.rsplit(".", 2)
    if len(all_splits) > 2:
        if ".".join(all_splits[1:]) in ("tar.gz", "tar.bz2"):
            return all_splits[0], ".".join([""] + all_splits[1:])
        return ".".join(all_splits[:-1]), "." + all_splits[2]
    if len(all_splits) == 2:
        return all_splits[0], "." + all_splits[1]
    if len(all_splits) == 1:
        return filename_with_ext, ""
    # Should never happen
    return "", ""


def conda_deps_to_pypi_deps(d: Dict[str, str]) -> Dict[str, str]:
    # This is a hack for now -- real fix would be to specify get_pinned_pypi_libs
    # but at least for now this allows channels to be specified in the conda_libs
    # but still work with pip
    r = {}  # type: Dict[str, str]
    for k, v in d.items():
        s = k.split("::")
        if len(s) > 1:
            package_name = s[1]
        else:
            package_name = k

        # HACK: ray package name differs between conda (ray-default) and pip (ray)
        # Convert ray-default to ray when using pip
        if package_name == "ray-default":
            package_name = "ray"

        r[package_name] = v
    return r


_UNDERSCORE_REGEX = re.compile(r"[-_.]+")


def normalize_to_underscore(name: str) -> str:
    return _UNDERSCORE_REGEX.sub("_", name).lower()


def channel_from_url(url: str) -> Optional[str]:
    up = urlparse(url)
    if up.hostname == "conda.anaconda.org":
        return up.path.split("/", 2)[1]
    return None


def channel_or_url(url: str) -> str:
    up = urlparse(url)
    if up.hostname == "conda.anaconda.org":
        return up.path.split("/", 2)[1]
    return url


def auth_from_urls(urls: List[str]) -> Optional[AuthBase]:
    auths_per_hostname = {
        cast(str, h): HTTPBasicAuth(username, pwd)
        for h, (username, pwd) in CONDA_SRCS_AUTH_INFO.items()
    }

    partial_auths = {}

    for url in urls:
        up = urlparse(url.strip("'\" \t\n"))
        if up.hostname and up.username:
            if up.password is not None:
                auths_per_hostname[up.hostname] = HTTPBasicAuth(
                    up.username, up.password
                )
            else:
                partial_auths[up.hostname] = up.username
    if auths_per_hostname or partial_auths:
        return PerHostAuth(auths_per_hostname, partial_auths)
    return None


class PerHostAuth(AuthBase):
    def __init__(
        self,
        auths_per_hostname: Mapping[str, AuthBase],
        partial_auths: Mapping[str, str],
    ):
        self._static_auths = auths_per_hostname
        self._partial_auths = partial_auths
        self._dynamic_auths = {}  # type: Dict[str, AuthBase]

        try:
            import keyring

            self._keyring = keyring
        except ImportError:
            self._keyring = None

    def __call__(self, r: PreparedRequest):
        h = self._get_hostname(r.url)
        if h:
            # Check if we have a static auth (username and password)
            auth_provider = self._static_auths.get(h or "<none>")
            if auth_provider:
                return auth_provider(r)

            # If we have a keyring, we can check for partial (just username) or full
            # auth from the keyring
            if self._keyring:
                if h in self._dynamic_auths:
                    r.register_hook("response", self._handle_401)
                    return self._dynamic_auths[h](r)
                elif h in self._partial_auths:
                    # Here we have a username but we don't have a password, we see
                    # if we can get it from the keyring
                    cred = self._get_and_update_credentials(h, self._partial_auths[h])
                    if cred is not None:
                        r.register_hook("response", self._handle_401)
                        return cred(r)
                else:
                    # We will suppose that other URLs can be looked up using keyring
                    # so on failure we will try to get the auth settings -- this means
                    # we only pay this cost if we fail first.
                    r.register_hook("response", self._handle_401)

        return r

    @staticmethod
    def _get_hostname(url: Optional[str]) -> Optional[str]:
        if url is None:
            return None
        up = urlparse(url)
        return up.hostname

    def _get_and_update_credentials(
        self, hostname: str, username: Optional[str] = None
    ) -> Optional[AuthBase]:
        if self._keyring is None:
            return None

        # Support keyring's get_credential interface which supports getting
        # credentials without a username. This is only available for
        # keyring>=15.2.0.
        if hasattr(self._keyring, "get_credential"):
            debug.conda_exec("Getting credentials from keyring for %s" % hostname)
            cred = self._keyring.get_credential(hostname, username)
            if cred is not None:
                debug.conda_exec(
                    "Successfully got keyring credentials for %s" % hostname
                )
                cred = HTTPBasicAuth(cred.username, cred.password)
                self._dynamic_auths[hostname] = cred
                return cred
            debug.conda_exec(
                "Attempted to get keyring credentials for %s but got nothing" % hostname
            )
            return None

        if username is not None:
            password = self._keyring.get_password(hostname, username)
            if password is not None:
                debug.conda_exec("Successfully got keyring password for %s" % hostname)
                cred = HTTPBasicAuth(username, password)
                self._dynamic_auths[hostname] = cred
                return cred
            debug.conda_exec(
                "Attempted to get keyring password for %s but got nothing" % hostname
            )
        return None

    def _handle_401(self, resp: Response, **kwargs: Any) -> Response:
        # This code is inspired from https://github.com/pypa/pip/blob/main/src/pip/_internal/network/auth.py

        # Only care about 401
        if resp.status_code != 401:
            return resp

        # If we are here, we know we have a keyring so we check for new credentials. We
        # also don't have a username to use.
        hostname = cast(str, self._get_hostname(resp.url))
        cred = self._get_and_update_credentials(hostname)

        if cred is None:
            # We couldn't get any credentials from keyring so we return the original
            # response (and error)
            return resp

        # Consume content and release the original connection to allow our new
        # request to reuse the same one.
        _ = resp.content
        resp.raw.release_conn()

        # Get a new request with authentication included
        req = cred(resp.request)
        req.deregister_hook("response", self._handle_401)

        with Session() as s:
            # Carry over cookies from the previous request if present
            if hasattr(resp, "cookies") and resp.cookies:
                s.cookies.update(resp.cookies)

            new_resp = s.send(req, **kwargs)
            new_resp.history.append(resp)
            return new_resp


def version_to_str(v: Version, overrides: Dict[str, Any]) -> str:
    # This code copies __str__ from Version basically but allows components to be
    # changed
    parts: List[str] = []

    # Epoch
    val = int(overrides.get("epoch", v.epoch))
    if v != 0:
        parts.append(f"{val}!")

    # Release segment
    val = overrides.get("release", v.release)
    parts.append(".".join(str(x) for x in cast(Tuple[int, ...], val)))

    # Pre-release
    val = overrides.get("pre", v.pre)
    if val is not None:
        parts.append("".join(str(x) for x in cast(Tuple[Union[str, int], ...], val)))

    # Post-release
    val = overrides.get("post", v.post)
    if val is not None:
        parts.append(f".post{val}")

    # Development release
    val = overrides.get("dev", v.dev)
    if val is not None:
        parts.append(f".dev{val}")

    # Local version segment
    val = overrides.get("local", v.local)
    if val is not None:
        parts.append(f"+{val}")

    return "".join(parts)


# HACK to prevent looking for python versions that do not exist on
# conda-forge.
version_maps = {
    "3.7.4": "3.7.5",
    "3.7.7": "3.7.8",
    "3.7.11": "3.7.12",
    "3.7.13": "3.7.12",
    "3.7.14": "3.7.12",
    "3.7.15": "3.7.12",
    "3.7.16": "3.7.12",
    "3.7.17": "3.7.12",
    "3.8.7": "3.8.8",
    "3.8.9": "3.8.10",
    "3.8.11": "3.8.12",
    "3.9.3": "3.9.4",
    "3.9.8": "3.9.9",
    "3.9.11": "3.9.12",
    "3.10.3": "3.10.4",
}


def sanitize_python_version(v: str) -> str:
    # Use major.minor version only to avoid OS-specific patch version constraints
    mapped_version = version_maps.get(v, v)
    version_parts = mapped_version.split('.')
    return f"{version_parts[0]}.{version_parts[1]}"


# Function heavily inspired from https://github.com/hauntsaninja/change_wheel_version
# MIT license of that source file:
# MIT License

# Copyright (c) 2023 hauntsaninja

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.


# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
def change_pypi_package_version(
    python_path: str, wheel_path: str, new_version_str: str
) -> str:
    base_dir, wheel_name = os.path.split(wheel_path)
    with tempfile.TemporaryDirectory() as build_dir:
        old_parts = parse_wheel_filename(wheel_name)

        distribution = wheel_name.split("-", 1)[0]
        old_version = old_parts[1]
        new_version = Version(new_version_str)
        new_version_str = str(new_version)

        old_slug = "%s-%s" % (distribution, old_version)
        new_slug = "%s-%s" % (distribution, new_version)

        subprocess.check_output(
            [python_path, "-m", "wheel", "unpack", "-d", build_dir, wheel_path]
        )
        if not os.path.isdir(os.path.join(build_dir, old_slug)) or not os.path.isdir(
            os.path.join(build_dir, old_slug, "%s.dist-info" % old_slug)
        ):
            raise ValueError("Invalid wheel expansion for '%s'" % wheel_path)

        # Copy everything
        shutil.move(
            os.path.join(build_dir, old_slug), os.path.join(build_dir, new_slug)
        )

        # Rename dist-info directory
        shutil.move(
            os.path.join(build_dir, new_slug, "%s.dist-info" % old_slug),
            os.path.join(build_dir, new_slug, "%s.dist-info" % new_slug),
        )

        # Rename data if it exists
        if os.path.exists(os.path.join(build_dir, new_slug, "%s.data" % old_slug)):
            shutil.move(
                os.path.join(build_dir, new_slug, "%s.data" % old_slug),
                os.path.join(build_dir, new_slug, "%s.data" % new_slug),
            )

        # Rewrite METADATA file
        metadata = os.path.join(
            build_dir, new_slug, "%s.dist-info" % new_slug, "METADATA"
        )

        # This is actually a non-conformant email policy as per
        # https://packaging.python.org/en/latest/specifications/core-metadata/
        # However, it works around this bug in setuptools in cases where the version is really long
        # https://github.com/pypa/setuptools/issues/3808
        max_line_length = 200
        policy = email.policy.compat32.clone(max_line_length=200)
        if len(new_version_str) >= max_line_length:
            raise ValueError(f"Version '%s' is too long" % new_version_str)

        with open(metadata, "rb") as f:
            parser = email.parser.BytesParser(policy=policy)
            msg = parser.parse(f)

        msg.replace_header("Version", new_version_str)
        with open(metadata, "wb") as f:
            f.write(msg.as_bytes())

        # Rewrites the RECORD file and generate the file.
        subprocess.check_output(
            [
                python_path,
                "-m",
                "wheel",
                "pack",
                "-d",
                build_dir,
                os.path.join(build_dir, new_slug),
            ]
        )

        # We check if we get the expected name
        expected_name = (
            reform_pypi_filename(distribution, new_version, old_parts[2], old_parts[3])
            + ".whl"
        )
        if os.path.isfile(os.path.join(build_dir, expected_name)):
            shutil.move(
                os.path.join(build_dir, expected_name),
                os.path.join(base_dir, expected_name),
            )
            os.unlink(wheel_path)
            return os.path.join(base_dir, expected_name)
        raise RuntimeError(
            "Could not rename wheel '%s'; expected '%s' got: %s"
            % (wheel_path, expected_name, ", ".join(os.listdir(build_dir)))
        )


def hash_directory(path: str) -> str:
    # The "hash" of a directory will be the last modified time and last modified
    # file. We could just use the time probably but this will hopefully inject
    # even more safety.
    last_modified_time = 0.0
    last_modified_file = None  # type: Optional[str]

    # Walk through all files and directories in the root directory
    top_level = True
    for directory, dirs, files in walk_without_cycles(path):
        if top_level:
            # Skip build and .egg-info since we don't want them to affect the
            # last modified file (it's not source)
            to_remove = []  # type: List[str]
            for d in dirs:
                if d == "build":
                    to_remove.append(d)
                if d.endswith(".egg-info"):
                    to_remove.append(d)
            for d in to_remove:
                dirs.remove(d)
            top_level = False

        for file in files:
            filename = os.path.join(directory, file)
            mod_time = os.path.getmtime(filename)
            # If this file is the most recently modified, update max_mod_time and max_file_dir
            if mod_time > last_modified_time:
                last_modified_time, last_modified_file = mod_time, filename

    debug.conda_exec(
        "Last modified file in '%s': '%s' at %s"
        % (path, last_modified_file, last_modified_time)
    )
    return hashlib.sha256(
        ("%s:%s" % (last_modified_time, last_modified_file)).encode()
    ).hexdigest()


def determine_uv_env_python_version(cwd: str) -> str:
    cmd_output, error_code = call_binary(
        args=["run", "python", "--version"],
        binary="uv",
        cwd=cwd,
        no_exception_on_error=True,
    )

    if error_code != 0:
        if (
            "which is incompatible with the project's Python requirement"
            in cmd_output.decode()
        ):
            raise EnvironmentError(
                "The uv environment's version is incompatible with the pyproject.toml requirements. "
                "Original error: %s" % cmd_output.decode()
            )
        else:
            raise EnvironmentError(
                "Could not call uv environment at '%s' to determine python version. "
                "uv run python --version returned error code %d with output='%s'"
                % (cwd, error_code, cmd_output.decode())
            )

    pattern = r"Python ([\d.]+)"

    match = re.search(pattern, cmd_output.decode())
    if match:
        return match.group(1)
    else:
        raise EnvironmentError(
            "Could not determine python version from uv environment at '%s'" % cwd
            + " uv run python --version output='%s'" % cmd_output.decode()
        )


def call_binary(
    args: List[str],
    binary: str,
    cwd: Optional[str] = None,
    no_exception_on_error: bool = False,
) -> Tuple[bytes, int]:
    try:
        debug.conda_exec("Binary call: %s" % str([binary] + args))
        return (
            subprocess.check_output(
                [binary] + args,
                stderr=subprocess.PIPE,
                cwd=cwd,
            ).strip(),
            0,
        )
    except subprocess.CalledProcessError as e:
        if no_exception_on_error:
            # The returned bytes: we simply concatenate stdout and stderr
            # here, as the downstream caller normally just matches a pattern
            # to determine the failure reason.
            return e.stdout.strip() + e.stderr.strip(), e.returncode
        else:
            raise CondaException(
                "Binary command for '{cmd}' returned error ({code}); "
                "see pretty-printed error above".format(cmd=e.cmd, code=e.returncode)
            ) from None


def get_best_compatible_packages(
    grouped_packages: Dict[str, List["PackageSpecification"]],
    supported_tags: List[Tag] = [],
) -> Dict[str, "PackageSpecification"]:

    best_packages = dict()  # type:  Dict[str, "PackageSpecification"]

    for key, package_list in grouped_packages.items():
        for t in supported_tags:
            # Tags are ordered from most-preferred to least preferred
            for p in package_list:
                full_filename = p.filename + p.url_format
                _, _, _, tags = parse_wheel_filename(full_filename)
                if t in tags:
                    best_packages[key] = p
                    debug.conda_exec("%s: matching package @ %s" % (key, p))
                    break
            else:
                # If we don't find a match, continue to next tag (and
                # skip break of outer loop on next line)
                continue
            break

    return best_packages


def compute_target_envs(
    target_env: Dict[str, str],
    python_versions: List[str],
    arch: str,
) -> List[Dict[str, str]]:
    # Strip out fields we don't support well.
    target_env.update(
        {
            "implementation_version": "",
            "python_full_version": "",
        }
    )

    if arch != arch_id():
        target_env.update(CONDA_SYS_MARKERS.get(arch, {}))

    # Workaround issue with platform_release not being clean enough for our packaged
    # version of packaging
    # Clean up platform_release to ensure it's in X[.Y[.Z]] format (digits, optional Y/Z, strip suffixes)
    pr = target_env.get("platform_release", "")
    m = re.match(r"(\d+)(?:\.(\d+))?(?:\.(\d+))?", pr)
    if m:
        cleaned = ".".join([group for group in m.groups(default="") if group != ""])
        target_env["platform_release"] = cleaned
    target_envs: List[Dict[str, str]] = []
    for v in python_versions:
        if v.count(".") != 2:
            new_target_env = target_env.copy()
            new_target_env.update({"python_version": v})
        else:
            new_target_env = target_env.copy()
            new_target_env.update(
                {
                    "python_full_version": v,
                    "python_version": ".".join(v.split(".")[:2]),
                    "implementation_version": v,
                }
            )
        target_envs.append(new_target_env)

    return target_envs


def evaluate_marker(
    marker: Optional[Marker], target_envs: List[Dict[str, str]]
) -> bool:
    if marker is None:
        return True  # No marker so we keep it
    try:
        keep_dep = marker.evaluate(target_envs[0])
        if any(marker.evaluate(env) != keep_dep for env in target_envs[1:]):
            raise InvalidEnvironmentException(
                f"Ambiguous package marker {marker}; "
                "inconsistent overlap with requested Python version. "
                "Restrict your Python version further to fully overlap "
                "with this marker."
            )
        return keep_dep

    except InvalidVersion as e:
        raise InvalidEnvironmentException(
            f"Package marker {marker} not supported"
        ) from e


def filter_packages_by_markers(
    packages: List["PackageSpecification"],
    python_full_version: str,
    arch: str,
    target_env: Optional[Dict[str, str]] = None,
) -> List["PackageSpecification"]:
    """
    Filters a list of packages based on their environment markers.
    More context in https://netflix.atlassian.net/browse/MLPMF-545.

    Parameters:

    target_env is a dictionary representing the target environment's properties,
    normally obtained by packaging.markers.default_environment(). A sample could be:
    {'implementation_name': 'cpython', 'implementation_version': '3.10.19',
     'os_name': 'posix', 'platform_machine': 'arm64', 'platform_release': '25.1.0',
     'platform_system': 'Darwin', 'platform_version':
     'Darwin Kernel Version 25.1.0: Mon Oct 20 19:34:05 PDT 2025; root:xnu-12377.41.6~2/RELEASE_ARM64_T6041',
     'python_full_version': '3.10.19', 'platform_python_implementation': 'CPython',
     'python_version': '3.10', 'sys_platform': 'darwin'}

    If target_env is not provided, then default_environment() will be called, and
    overrides will be applied to reflect information provided in python_full_version
    and arch.

    arch is our own donation defined in arch_id(). We have our own overrides
    of target_env based on the arch provided.
    """
    if python_full_version.count(".") != 2:
        raise InvalidEnvironmentException(
            "Full python version expected to be known "
            "-- this is a bug. Please report it"
        )
    target_envs = compute_target_envs(
        target_env or default_environment(),
        [python_full_version],
        arch,
    )

    # We now filter based on the markers -- if we try to filter based on something
    # that is empty, an error (packaging.version.InvalidVersion) is raised so we
    # can catch that.
    filtered_packages: List["PackageSpecification"] = []

    for pkg in packages:
        if not pkg.environment_marker:
            filtered_packages.append(pkg)
        else:
            marker = Marker(pkg.environment_marker)
            if evaluate_marker(marker, target_envs):
                filtered_packages.append(pkg)
            else:
                debug.conda_exec(
                    "Skipping dependency '%s' due to marker mismatch" % pkg.filename
                )

    return filtered_packages


def filter_user_reqs_by_markers(
    deps: Dict[str, List[str]],
    python_version: str,
    arch: str,
    target_env: Optional[Dict[str, str]] = None,
) -> Dict[str, List[str]]:
    """
    Filters out user specified dependencies based on markers.

    Parameters
    ----------
    deps : Dict[str, List[str]]
        The dependencies to filter
    python_version : str
        The python version to filter against
    arch : str
        The architecture to filter against
    target_env : Optional[Dict[str, str]]
        The target environment to filter against

    Returns
    -------
    Dict[str, List[str]]
        The filtered dependencies
    """
    # This is in part to support markers for conda (and some other cases).
    # To do so, we need to manually filter
    # out dependencies. We may not know specifically, at this time, what version of
    # python we are going to be using and if markers filter on python versions, that
    # may be a problem if for some choice of python we need to keep the dependency and
    # for others we don't. As an example, if we have:
    #  - python_version >= 3.10
    #  - marker: python_version <= 3.11
    # that would not work because we don't know if we are getting 3.10, 3.11 or 3.12 at
    # that time.
    # We only deal with major.minor for now as a "heuristic". It is likely that
    # it won't matter much. If this is a problem, we need to add patch version too.
    # Consider:
    #  - python_version >= 3.10,<3.11
    #  - marker: python_version <= 3.10.5
    # We would not detect this possible problem and include the dependency
    python_req = get_requirement(python_version)

    acceptable_versions = list(python_req.specifier.filter(_generate_python_versions()))
    if not acceptable_versions:
        # Specific version (this is best)
        python_req_str = str(python_req.specifier)
        if not python_req_str.startswith("=="):
            raise InvalidEnvironmentException(
                f"Python version {python_version} does not match any known version"
            )
        acceptable_versions = [python_req_str[2:]]

    target_envs = compute_target_envs(
        target_env or default_environment(),
        acceptable_versions,
        arch,
    )
    # We now filter based on the markers -- if we try to filter based on something
    # that is empty, an error (packaging.version.InvalidVersion) is raised so we
    # can catch that.
    new_deps = {}  # type: Dict[str, List[str]]
    for dep_type, dep_list in deps.items():
        new_deps[dep_type] = []
        if dep_type not in ("conda", "pypi"):
            new_deps[dep_type] = dep_list
            continue
        for dep in dep_list:
            if ";" not in dep:
                # There is definitely no marker so we can keep the dependency as is.
                # This speeds things up and also avoids issues with conda packages that
                # do not meet the Requirement syntax (like ones starting with _)
                new_deps[dep_type].append(dep)
                continue

            req = get_requirement(dep)
            if evaluate_marker(req.marker, target_envs):
                new_deps[dep_type].append(dep.split(";")[0])  # Remove the marker part
            else:
                debug.conda_exec(
                    "Skipping dependency '%s' due to marker mismatch"
                    % dep.split(";")[0]
                )
                continue
    return new_deps


def compute_file_hash(file_path: str) -> str:
    if not file_path:
        return ""

    try:
        with open(file_path, "rb") as f:
            file_content_hash = hashlib.sha256(f.read()).hexdigest()
            return file_content_hash
    except (IOError, OSError):
        raise IOError(f"Could not read file '{file_path}' for hashing") from None


class WithDir:
    # WARNING: os.chdir is not compatible with thread processing so do not use in
    # a context where multiple threads can exist.
    def __init__(self, new_dir: str):
        self._current_dir = os.getcwd()
        self._new_dir = new_dir

    def __enter__(self):
        os.chdir(self._new_dir)
        return self._new_dir

    def __exit__(self, exc_type, exc_val, exc_tb):
        os.chdir(self._current_dir)



# pyright: strict, reportTypeCommentUsage=false, reportMissingTypeStubs=false
import copy

from itertools import chain
from typing import Any, Callable, ClassVar, Dict, List, Optional, cast

from metaflow.metaflow_environment import InvalidEnvironmentException
from metaflow.unbounded_foreach import UBF_CONTROL, UBF_TASK

from metaflow._vendor.packaging.utils import canonicalize_version

from .utils import dict_to_strlist, merge_dep_dicts


class StepRequirementIface:
    @property
    def from_name(self) -> Optional[str]:
        return None

    @property
    def from_pathspec(self) -> Optional[str]:
        return None

    @property
    def python(self) -> Optional[str]:
        return None

    @property
    def is_fetch_at_exec(self) -> Optional[bool]:
        return None

    @property
    def is_disabled(self) -> Optional[bool]:
        return None

    @property
    def packages(self) -> Dict[str, Dict[str, str]]:
        return {}

    @property
    def sources(self) -> Dict[str, List[str]]:
        return {}

    @property
    def file_paths(self) -> Dict[str, List[str]]:
        return {}

    @property
    def full_id_unique_keys(self) -> Dict[str, str]:
        return {}

    @property
    def extras(self) -> Dict[str, List[str]]:
        return {}

    def default_disabled(self, ubf_context: str) -> Optional[bool]:
        return None


class StepRequirementMixin(StepRequirementIface):
    defaults = {
        "disabled": None,
    }  # type: Dict[str, Any]

    _derived_classes_fullnames: ClassVar[Dict[str, str]] = {}

    @classmethod
    def get_derived_classes_fullnames(cls) -> Dict[str, str]:
        return cls._derived_classes_fullnames

    def __init_subclass__(cls, *args, **kwargs):
        super().__init_subclass__(*args, **kwargs)
        # using __name__ not __qualname__ to align with
        # MutableStep.decorator_specs
        # Only track the ones that the user will use (with names)
        cls_user_name: Optional[str] = getattr(cls, "name", None)
        if cls_user_name is not None:
            StepRequirementMixin._derived_classes_fullnames[
                f"{cls.__module__}.{cls.__name__}"
            ] = cls_user_name

    @property
    def is_disabled(self) -> Optional[bool]:
        if self.attributes["disabled"] is None:  # type: ignore[attr-defined]
            if (
                self.python
                or self.from_name
                or self.from_pathspec
                or self.packages
                or self.sources
                or self.file_paths
            ):
                return False
            return None
        return self.attributes["disabled"]  # type: ignore[attr-defined]


class StepRequirement(StepRequirementIface):
    def __init__(self) -> None:
        self._name = None  # type: Optional[str]
        self._pathspec = None  # type: Optional[str]
        self._python = None  # type: Optional[str]
        self._fetch_at_exec = None  # type: Optional[bool]
        self._disabled = None  # type: Optional[bool]
        self._packages = {}  # type: Dict[str, Dict[str, str]]
        self._sources = {}  # type: Dict[str, List[str]]
        self._file_paths = {}  # type: Dict[str, List[str]]
        self._extras = {}  # type: Dict[str, List[str]]
        self._full_id_unique_keys = {}  # type: Dict[str, str]
        self._default_disabled = {
            UBF_CONTROL: None,
            UBF_TASK: None,
        }  # type: Dict[str, Optional[bool]]

    def copy(self) -> "StepRequirement":
        n = StepRequirement()
        n._name = self._name
        n._pathspec = self._pathspec
        n._python = self._python
        n._fetch_at_exec = self._fetch_at_exec
        n._disabled = self._disabled
        n._packages = copy.deepcopy(self._packages)
        n._sources = copy.deepcopy(self._sources)
        n._file_paths = copy.deepcopy(self._file_paths)
        n._full_id_unique_keys = copy.deepcopy(self._full_id_unique_keys)
        n._extras = copy.deepcopy(self._extras)
        n._default_disabled = copy.deepcopy(self._default_disabled)
        return n

    @property
    def from_name(self) -> Optional[str]:
        return self._name

    @property
    def from_pathspec(self) -> Optional[str]:
        return self._pathspec

    @property
    def from_env_name(self) -> Optional[str]:
        if self.from_name:
            return self.from_name
        elif self.from_pathspec:
            return "step:%s" % self.from_pathspec
        return None

    @property
    def python(self) -> Optional[str]:
        return self._python

    @python.setter
    def python(self, value: str):
        self._python = value

    @property
    def is_fetch_at_exec(self) -> Optional[bool]:
        return self._fetch_at_exec

    @property
    def is_disabled(self) -> Optional[bool]:
        if self._disabled is None:
            if (
                self.python
                or self.from_name
                or self.from_pathspec
                or self.packages
                or self.sources
                or self.is_fetch_at_exec
            ):
                return False
            return None
        return self._disabled

    @property
    def packages(self) -> Dict[str, Dict[str, str]]:
        return copy.deepcopy(self._packages)

    @property
    def packages_as_str(self) -> Dict[str, List[str]]:
        result = {}  # type: Dict[str, List[str]]
        for category, values in self.packages.items():
            result[category] = dict_to_strlist(values)
        return result

    @packages.setter  # type: ignore[no-redef,attr-defined]
    def packages(self, value: Dict[str, Dict[str, str]]):
        self._packages = value

    @property
    def sources(self) -> Dict[str, List[str]]:
        return copy.deepcopy(self._sources)

    @sources.setter
    def sources(self, value: Dict[str, List[str]]):
        self._sources = value

    @property
    def file_paths(self) -> Dict[str, List[str]]:
        return copy.deepcopy(self._file_paths)

    @file_paths.setter
    def file_paths(self, value: Dict[str, List[str]]):
        self._file_paths = value

    @property
    def full_id_unique_keys(self) -> Dict[str, str]:
        return copy.deepcopy(self._full_id_unique_keys)

    @full_id_unique_keys.setter
    def full_id_unique_keys(self, value: Dict[str, str]):
        self._full_id_unique_keys = value

    @property
    def extras(self) -> Dict[str, List[str]]:
        return copy.deepcopy(self._extras)

    @extras.setter
    def extras(self, value: Dict[str, List[str]]):
        self._extras = value

    def default_disabled(self, ubf_context: str) -> Optional[bool]:
        return self._default_disabled[ubf_context]

    def __repr__(self) -> str:
        disabled_part = (
            "disabled=%s" % self.is_disabled if self.is_disabled is not None else ""
        )
        if self.is_disabled:
            return "StepReq[%s]" % disabled_part
        python_part = "python=%s" % self.python if self.python else ""
        fetch_at_exec_part = (
            "fetch_at_exec=%s" % self.is_fetch_at_exec
            if self.is_fetch_at_exec is not None
            else ""
        )
        from_env_part = "from=%s" % self.from_env_name if self.from_env_name else ""
        packages_part = "packages=%s" % self.packages if self.packages else ""
        sources_part = "sources=%s" % self.sources if self.sources else ""
        extras_part = "extras=%s" % self.extras if self.extras else ""
        return "StepReq[%s]" % "; ".join(
            filter(
                lambda x: x,
                [
                    disabled_part,
                    python_part,
                    fetch_at_exec_part,
                    from_env_part,
                    packages_part,
                    sources_part,
                    extras_part,
                ],
            )
        )

    def merge_update(self, other: StepRequirementIface):
        def _check_and_return(f: str, v1: Any, v2: Any) -> Any:
            if v1 is not None:
                if v2 is None or v1 == v2:
                    return v1
                raise InvalidEnvironmentException(
                    "%s is specified with incompatible values: %s and %s" % (f, v1, v2)
                )
            else:
                return v2

        self._internal_merge(other, _check_and_return, merge_dep_dicts)

    def override_update(self, other: StepRequirementIface):
        def _override_if_none(f: str, v1: Any, v2: Any) -> Any:
            if v2 is None:
                return v1
            return v2

        def _update_dict(d1: Dict[str, str], d2: Dict[str, str]) -> Dict[str, str]:
            d1.update(d2)
            return d1

        self._internal_merge(other, _override_if_none, _update_dict)

    def _internal_merge(
        self,
        other: StepRequirementIface,
        check_func: Callable[[str, Any, Any], Any],
        merge_func: Callable[[Dict[str, str], Dict[str, str]], Dict[str, str]],
    ):
        self._name = check_func("name", self.from_name, other.from_name)
        self._pathspec = check_func("pathspec", self.from_pathspec, other.from_pathspec)
        self._python = check_func("python", self.python, other.python)
        self._fetch_at_exec = check_func(
            "fetch_at_exec", self.is_fetch_at_exec, other.is_fetch_at_exec
        )
        self._disabled = check_func("disabled", self.is_disabled, other.is_disabled)
        for ubf_context in (UBF_CONTROL, UBF_TASK):
            self._default_disabled[ubf_context] = check_func(
                "default_disabled",
                self.default_disabled(ubf_context),
                other.default_disabled(ubf_context),
            )

        other_packages = other.packages

        for category, packages in other_packages.items():
            self._packages[category] = merge_func(
                self._packages.get(category, {}), packages
            )

        other_sources = other.sources
        for category, sources in other_sources.items():
            self._sources[category] = list(
                set(sources + self._sources.get(category, []))
            )

        other_file_paths = other.file_paths
        for category, paths in other_file_paths.items():
            self._file_paths[category] = list(
                set(paths + self._file_paths.get(category, []))
            )

        other_extras = other.extras
        for category, extras in other_extras.items():
            self._extras.setdefault(category, []).extend(extras or [])

        other_full_id_unique_keys = other.full_id_unique_keys
        for key, value in other_full_id_unique_keys.items():
            # We use override rather than list extension here for same keys,
            # as we assume it rare that different decorators will have same
            # full_id_unique_keys.
            self._full_id_unique_keys[key] = value

        # Special handling for pathspec/name
        if other.from_name is not None and other.from_pathspec is not None:
            raise InvalidEnvironmentException(
                "Cannot specify both `name` and `pathspec` in a decorator"
            )
        if other.from_name is not None:
            self._pathspec = None  # Use the latest and override any previous one
        elif other.from_pathspec is not None:
            self._name = None


class CondaRequirementDecoratorMixin(StepRequirementMixin):
    defaults = {
        "python": None,
        "libraries": {},
        "channels": [],
        # The next fields are deprecated in favor of @named_env and @pypi
        "pip_packages": {},
        "pip_sources": [],
        "name": None,
        "pathspec": None,
        "fetch_at_exec": None,
        **StepRequirementMixin.defaults,
    }

    @property
    def python(self) -> Optional[str]:
        return self.attributes["python"]  # type: ignore[attr-defined]

    @property
    def from_name(self) -> Optional[str]:
        return self.attributes["name"]  # type: ignore[attr-defined]

    @property
    def pathspec(self) -> Optional[str]:
        return self.attributes["pathspec"]  # type: ignore[attr-defined]

    @property
    def is_fetch_at_exec(self) -> Optional[bool]:
        return self.attributes["fetch_at_exec"]  # type: ignore[attr-defined]

    @property
    def packages(self) -> Dict[str, Dict[str, str]]:
        return {
            "conda": {
                k: v
                for k, v in cast(Dict[str, str], self.attributes["libraries"]).items()  # type: ignore[attr-defined]
            },
            "pypi": {
                k: canonicalize_version(v)
                for k, v in cast(
                    Dict[str, str], self.attributes["pip_packages"]  # type: ignore[attr-defined]
                ).items()
            },
        }

    @property
    def sources(self) -> Dict[str, List[str]]:
        return {
            "conda": [k for k in cast(List[str], self.attributes["channels"])],  # type: ignore[attr-defined]
            "pypi": [k for k in cast(List[str], self.attributes["pip_sources"])],  # type: ignore[attr-defined]
        }


class PypiRequirementDecoratorMixin(StepRequirementMixin):
    defaults = {
        "python": None,
        "packages": {},
        "conda_only": {},
        "extra_indices": [],
        "extras": [],
        # The next fields are deprecated in favor of @named_env
        "sources": [],
        "name": None,
        "pathspec": None,
        "fetch_at_exec": None,
        **StepRequirementMixin.defaults,
    }

    @property
    def python(self) -> Optional[str]:
        return self.attributes["python"]  # type: ignore[attr-defined]

    @property
    def from_name(self) -> Optional[str]:
        return self.attributes["name"]  # type: ignore[attr-defined]

    @property
    def pathspec(self) -> Optional[str]:
        return self.attributes["pathspec"]  # type: ignore[attr-defined]

    @property
    def is_fetch_at_exec(self) -> Optional[bool]:
        return self.attributes["fetch_at_exec"]  # type: ignore[attr-defined]

    @property
    def packages(self) -> Dict[str, Dict[str, str]]:
        return {
            "pypi": {
                k: canonicalize_version(v)
                for k, v in cast(Dict[str, str], self.attributes["packages"]).items()  # type: ignore[attr-defined]
            },
            "npconda": {
                k: canonicalize_version(v)
                for k, v in cast(Dict[str, str], self.attributes["conda_only"]).items()  # type: ignore[attr-defined]
            },
        }

    @property
    def sources(self) -> Dict[str, List[str]]:
        return {
            "pypi": [
                k
                for k in cast(
                    List[str],
                    chain(self.attributes["sources"], self.attributes["extra_indices"]),  # type: ignore[attr-defined]
                )
            ]
        }

    @property
    def extras(self) -> Dict[str, List[str]]:
        return {"pypi": [f"--{extra}" for extra in self.attributes["extras"]]}  # type: ignore[attr-defined]


class NamedEnvRequirementDecoratorMixin(StepRequirementMixin):
    defaults = {
        "name": None,
        "pathspec": None,
        "fetch_at_exec": None,
        **StepRequirementMixin.defaults,
    }

    @property
    def from_name(self) -> Optional[str]:
        return self.attributes["name"]  # type: ignore[attr-defined]

    @property
    def from_pathspec(self) -> Optional[str]:
        return self.attributes["pathspec"]  # type: ignore[attr-defined]

    @property
    def is_fetch_at_exec(self) -> Optional[bool]:
        return self.attributes["fetch_at_exec"]  # type: ignore[attr-defined]


class SysPackagesRequirementDecoratorMixin(StepRequirementMixin):
    defaults = {
        "packages": None,
        **StepRequirementMixin.defaults,
    }

    @property
    def packages(self) -> Dict[str, Dict[str, str]]:
        return {
            "sys": {
                k: v
                for k, v in cast(Dict[str, str], self.attributes["packages"]).items()  # type: ignore[attr-defined]
            }
        }



import os
from metaflow.metaflow_config import (
    DATASTORE_SYSROOT_S3,
    DATASTORE_SYSROOT_AZURE,
    DATASTORE_SYSROOT_GS,
    DATASTORE_SYSROOT_LOCAL,
)
from metaflow.metaflow_config_funcs import from_conf, get_validate_choice_fn


# Set to true if running the tests with a local datastore
CONDA_TEST = from_conf("CONDA_TEST", False)

# Datastores for which caching should be skipped
CONDA_IGNORE_CACHING_DATASTORES = from_conf(
    "CONDA_IGNORE_CACHING_DATASTORES", []
)

# HACK -- work around an issue with micromamba where using a channel_alias
# causes the packages to be considered invalid (URL verification). This
# value should be set to the channel_alias prefix that will be replaced by
# conda.anaconda.org. You should most likely not need this.
CONDA_HACK_CHANNEL_ALIAS = from_conf(
    "CONDA_HACK_CHANNEL_ALIAS",
    None,
)

CONDA_S3ROOT = from_conf(
    "CONDA_S3ROOT",
    os.path.join(DATASTORE_SYSROOT_S3, "conda_env") if DATASTORE_SYSROOT_S3 else None,
)

CONDA_AZUREROOT = from_conf(
    "CONDA_AZUREROOT",
    (
        os.path.join(DATASTORE_SYSROOT_AZURE, "conda_env")
        if DATASTORE_SYSROOT_AZURE
        else None
    ),
)

CONDA_GSROOT = from_conf(
    "CONDA_GSROOT",
    os.path.join(DATASTORE_SYSROOT_GS, "conda_env") if DATASTORE_SYSROOT_GS else None,
)

CONDA_LOCALROOT = from_conf(
    "CONDA_LOCALROOT",
    (
        os.path.join(DATASTORE_SYSROOT_LOCAL, "conda_env")
        if DATASTORE_SYSROOT_LOCAL
        else None
    ),
)

CONDA_MAGIC_FILE_V2 = from_conf("CONDA_MAGIC_FILE_V2", "conda_v2.cnd")

# Use an alternate dependency resolver for conda packages instead of conda
# Mamba promises faster package dependency resolution times, which
# should result in an appreciable speedup in flow environment initialization.
CONDA_DEPENDENCY_RESOLVER = from_conf(
    "CONDA_DEPENDENCY_RESOLVER",
    "mamba",
    get_validate_choice_fn(["mamba", "conda", "micromamba"]),
)

# For pure PYPI environments, if you want to support those, set to the pypi resolver.
# Set to "none" if you do not want to support this functionality.
CONDA_PYPI_DEPENDENCY_RESOLVER = from_conf(
    "CONDA_PYPI_DEPENDENCY_RESOLVER",
    "uv",
    get_validate_choice_fn(["pip", "uv", "none"]),
)

# For mixed conda/pypi environments, if you want to support those, set this to 'conda-lock'
# Set to "none" if you want to disable this functionality.
CONDA_MIXED_DEPENDENCY_RESOLVER = from_conf(
    "CONDA_MIXED_DEPENDENCY_RESOLVER",
    "conda-lock",
    get_validate_choice_fn(["conda-lock", "none"]),
)

# Timeout trying to acquire the lock to create environments
CONDA_LOCK_TIMEOUT = from_conf("CONDA_LOCK_TIMEOUT", 3600)

# Location within CONDA_<DS>ROOT of the packages directory
CONDA_PACKAGES_DIRNAME = from_conf("CONDA_PACKAGES_DIRNAME", "packages")
# Ditto for the envs directory
CONDA_ENVS_DIRNAME = from_conf("CONDA_ENVS_DIRNAME", "envs")


# CONDA_REMOTE_INSTALLER_DIRNAME = from_conf(
#     "CONDA_REMOTE_INSTALLER_DIRNAME", "conda-remote"
# )
CONDA_REMOTE_INSTALLER_DIRNAME = from_conf("CONDA_REMOTE_INSTALLER_DIRNAME")

# Binary within CONDA_REMOTE_INSTALLER_DIRNAME to use as the binary on remote instances.
# Use {arch} to specify the architecture. This should be fully functional binary
# CONDA_REMOTE_INSTALLER = from_conf("CONDA_REMOTE_INSTALLER", "conda-{arch}")
CONDA_REMOTE_INSTALLER = from_conf("CONDA_REMOTE_INSTALLER")


# Directory in CONDA_*ROOT that contains the local distribution tarballs. If not specified,
# a conda executable needs to be installed and will be used (if locatable in PATH)
# CONDA_LOCAL_DIST_DIRNAME = from_conf("CONDA_LOCAL_DIST_DIRNAME", "conda-local")
CONDA_LOCAL_DIST_DIRNAME = from_conf("CONDA_LOCAL_DIST_DIRNAME")

# Tar-ball containing the local distribution of conda to use.
# CONDA_LOCAL_DIST = from_conf("CONDA_LOCAL_DIST", "conda-{arch}.tgz")
CONDA_LOCAL_DIST = from_conf("CONDA_LOCAL_DIST")

# Path to the local conda distribution. If this is specified and a conda distribution
# does not exist at this path, Metaflow will attempt to install it using
# CONDA_LOCAL_DIST_DIRNAME and CONDA_LOCAL_DIST
# CONDA_LOCAL_PATH = from_conf("CONDA_LOCAL_PATH", "/usr/local/libexec/metaflow-conda")
CONDA_LOCAL_PATH = from_conf("CONDA_LOCAL_PATH")

# Preferred Format for Conda packages
CONDA_PREFERRED_FORMAT = from_conf(
    "CONDA_PREFERRED_FORMAT",
    "none",
    get_validate_choice_fn([".tar.bz2", ".conda", "none"]),
)

CONDA_DEFAULT_PYPI_SOURCE = from_conf("CONDA_DEFAULT_PYPI_SOURCE", None)

# Allows you to specify URLs that need to be authenticated (basically Conda or PYPI
# package sources). Metaflow will glean all the information it can from traditional
# configuration sources (like pip.conf) but this allows you to specify additional URLs
# for which authentication is required. A dictionary should be passed in (or a JSON
# version of it in string form).
# key: hostname that needs authentication
# value: a tuple/list of username and password
CONDA_SRCS_AUTH_INFO = from_conf("CONDA_SRCS_AUTH_INFO", {})

PYLOCK_TOML_DEPENDENCY_RESOLVER = from_conf(
    "PYLOCK_TOML_DEPENDENCY_RESOLVER",
    "pylock_toml",
    get_validate_choice_fn(["pylock_toml"]),
)

CONDA_REMOTE_COMMANDS = ("batch", "kubernetes")

# List of system dependencies that are allowed to indicate the system to build on
CONDA_SYS_DEPENDENCIES = ("__cuda", "__glibc")

# Default system dependencies when not specified. Note that the `linux-64` defaults are
# used as default when building on the remote platform.
#
# As an example, you can set it to:
# CONDA_SYS_DEFAULT_PACKAGES = {
#     "linux-64": {"__glibc": os.environ.get("CONDA_OVERRIDE_GLIBC", "2.27")},
# }
CONDA_SYS_DEFAULT_PACKAGES = from_conf(
    "CONDA_SYS_DEFAULT_PACKAGES",
    {
        "linux-64": {"__glibc": os.environ.get("CONDA_OVERRIDE_GLIBC", "2.27")},
    },
)

# Packages to add when building for GPU machines (ie: if there is a GPU resource
# requirement). As an example you can set this to:
# CONDA_SYS_DEFAULT_GPU_PACKAGES = {
#     "__cuda": os.environ.get("CONDA_OVERRIDE_CUDA", "11.8=0")
# }
CONDA_SYS_DEFAULT_GPU_PACKAGES = {}

# Markers to assume for different architectures
CONDA_SYS_MARKERS = from_conf("CONDA_SYS_MARKERS", {})

# Packages that are needed when creating an intermediate conda environment to resolve
# a pypi environment. This can be used to add packages needed for authentication for
# example.
builder_packages = []
if os.environ.get("GOOGLE_APPLICATION_CREDENTIALS"):
    builder_packages = ["keyrings.google-artifactregistry-auth"]

CONDA_BUILDER_ENV_PACKAGES = from_conf("CONDA_BUILDER_ENV_PACKAGES", builder_packages)

# All architectures used by Conda
CONDA_ALL_ARCHS = from_conf("CONDA_ALL_ARCHS", ["linux-64", "osx-64", "osx-arm64"])


def _validate_remote_latest(name, value):
    if not value:
        raise MetaflowException("%s must not be empty." % name)
    if value[0] == ":" and value not in (":none:", ":username:", ":any:"):
        raise MetaflowException(
            "%s can only have special values ':none:', ':username:' or ':any:'" % name
        )


# Set to:
#  - :none: if default environments should never be looked up remotely
#  - :username: if default environments should be fetched remotely picking the latest
#    environment by the user
#  - :any: if default environments should be fetched remotely picking the latest
#    environment
#  - comma separated list of names: same as :username: but with a group of users.
CONDA_USE_REMOTE_LATEST = from_conf(
    "CONDA_USE_REMOTE_LATEST", ":none:", validate_fn=_validate_remote_latest
)

DEBUG_OPTIONS = ["conda"]



# pyright: strict, reportTypeCommentUsage=false, reportMissingTypeStubs=false

from typing import ClassVar, Dict
from metaflow.decorators import FlowDecorator
from metaflow.metaflow_environment import InvalidEnvironmentException

from .conda_common_decorator import (
    CondaRequirementDecoratorMixin,
    NamedEnvRequirementDecoratorMixin,
    SysPackagesRequirementDecoratorMixin,
    PypiRequirementDecoratorMixin,
)


class PackageRequirementFlowDecorator(FlowDecorator):
    name = "package_base"

    allow_multiple = True
    _derived_classes_fullnames: ClassVar[Dict[str, str]] = {}

    @classmethod
    def get_derived_classes_fullnames(cls) -> Dict[str, str]:
        return cls._derived_classes_fullnames

    def __init_subclass__(cls, *args, **kwargs):
        super().__init_subclass__(*args, **kwargs)
        # using __name__ not __qualname__ to align with
        # MutableFlow.decorator_specs
        # Only track the ones that the user will use (with names)
        cls_user_name = getattr(cls, "name", None)
        if cls_user_name is not None:
            PackageRequirementFlowDecorator._derived_classes_fullnames[
                f"{cls.__module__}.{cls.__name__}"
            ] = cls_user_name

    def flow_init(
        self, flow, graph, environment, flow_datastore, metadata, logger, echo, options
    ):
        if environment.TYPE != "conda":
            raise InvalidEnvironmentException(
                "The *%s* decorator requires " "--environment=conda" % self.name
            )


class CondaRequirementFlowDecorator(
    CondaRequirementDecoratorMixin, PackageRequirementFlowDecorator
):
    """
    Specifies the Conda packages for all steps of the flow.

    Use `@conda_base`, `@pypi_base` or `@named_env_base` to set common libraries
    required by all steps and use `@conda` or `@pypi` to specify step-specific additions.

    Parameters
    ----------
    libraries : Dict[str, str], default {}
        Libraries to use for this flow. The key is the name of the package
        and the value is the version to use. Note that versions can
        be specified either as a specific version or as a comma separated string
        of constraints like "<2.0,>=1.5".
    channels : List[str], default []
        Additional channels to search
    python : str, optional, default None
        Version of Python to use, e.g. '3.7.4'. If not specified, the current version
        will be used.
    disabled : bool, default False
        If set to True, uses the external environment.
    name : str, optional, default None
        DEPRECATED -- use `@named_env(name=)` instead.
        If specified, can refer to a named environment. The environment referred to
        here will be the one used for this flow. If specified, nothing else can be
        specified in this decorator. In the name, you can use `@{}` values and
        environment variables will be used to substitute.
    pathspec : str, optional, default None
        DEPRECATED -- use `@named_env(pathspec=)` instead.
        If specified, can refer to the pathspec of an existing step. The environment
        of this referred step will be used here. If specified, nothing else can be
        specified in this decorator. In the pathspec, you can use `@{}` values and
        environment variables will be used to substitute.
    pip_packages : Dict[str, str], default {}
        DEPRECATED -- use `@pypi(packages=)` instead.
        Same as libraries but for pip packages.
    pip_sources : List[str], default []
        DEPRECATED -- use `@pypi(extra_indices=)` instead.
        Same as channels but for pip sources.
    fetch_at_exec : bool, default False
        DEPRECATED -- use `@named_env(fetch_at_exec=)` instead.
        If set to True, the environment will be fetched when the task is
        executing as opposed to at the beginning of the flow (or at deploy time if
        deploying to a scheduler). This option requires name or pathspec to be
        specified. This is useful, for example, if you want this flow to always use
        the latest named environment when it runs as opposed to the latest when it
        is deployed.
    """

    name = "conda_base"

    def flow_init(
        self, flow, graph, environment, flow_datastore, metadata, logger, echo, options
    ):
        deprecated_keys = set(
            ("pip_packages", "pip_sources", "fetch_at_exec", "name", "pathspec")
        ).intersection((k for k, v in self.attributes.items() if v))

        if deprecated_keys:
            echo(
                "*DEPRECATED*: Using '%s' in '@%s' is deprecated. Please use '@pypi_base' "
                "or '@named_env_base' instead. "
                % (", ".join(deprecated_keys), self.name)
            )
        super().flow_init(
            flow, graph, environment, flow_datastore, metadata, logger, echo, options
        )


class PypiRequirementFlowDecorator(
    PypiRequirementDecoratorMixin, PackageRequirementFlowDecorator
):
    """
    Specifies the Pypi packages for all steps of the flow.

    Use `@conda_base`, `@pypi_base` or `@named_env_base` to set common libraries
    required by all steps and use `@conda` or `@pypi` to specify step-specific additions.

    Parameters
    ----------
    packages : Dict[str, str], default {}
        Packages to use for this flow. The key is the name of the package
        and the value is the version to use.
    conda_only: Dict[str, str], default {}
        Packages that are only available in conda. These packages should not depend on
        any python package (example ffmpeg). Note that this is different than using
        the `@conda_base` decorator as using `conda_only` will still build a Pypi
        environment instead of a mixed one.
    extra_indices : List[str], default []
        Additional sources to search for
    extras: List[str], default []
        Extra arguments to pass to the resolver. For example "pre" to include
        pre-release packages.
    python : str, optional, default None
        Version of Python to use, e.g. '3.7.4'. If not specified, the current python
        version will be used.
    disabled : bool, default False
        If set to True, uses the external environment.
    name : str, optional, default None
        DEPRECATED -- use `@named_env(name=)` instead.
        If specified, can refer to a named environment. The environment referred to
        here will be the one used for this flow. If specified, nothing else can be
        specified in this decorator. In the name, you can use `@{}` values and
        environment variables will be used to substitute.
    pathspec : str, optional, default None
        DEPRECATED -- use `@named_env(name=)` instead.
        If specified, can refer to the pathspec of an existing step. The environment
        of this referred step will be used here. If specified, nothing else can be
        specified in this decorator. In the name, you can use `@{}` values and
        environment variables will be used to substitute.
    fetch_at_exec : bool, default False
        DEPRECATED -- use `@named_env(name=)` instead.
        If set to True, the environment will be fetched when the task is
        executing as opposed to at the beginning of the flow (or at deploy time if
        deploying to a scheduler). This option requires name or pathspec to be
        specified. This is useful, for example, if you want this flow to always use
        the latest named environment when it runs as opposed to the latest when it
        is deployed.
    """

    name = "pypi_base"

    def flow_init(
        self, flow, graph, environment, flow_datastore, metadata, logger, echo, options
    ):
        deprecated_keys = set(
            ("sources", "fetch_at_exec", "name", "pathspec")
        ).intersection((k for k, v in self.attributes.items() if v))

        if deprecated_keys:
            echo(
                "*DEPRECATED*: Using '%s' in '@%s' is deprecated. Please use "
                "'@named_env_base' instead. " % (", ".join(deprecated_keys), self.name)
            )

        return super().flow_init(
            flow, graph, environment, flow_datastore, metadata, logger, echo, options
        )


class NamedEnvRequirementFlowDecorator(
    NamedEnvRequirementDecoratorMixin, PackageRequirementFlowDecorator
):
    """
    Specifies a named environment to extract the environment from

    Use `@conda_base`, `@pypi_base` or `@named_env_base` to set common libraries
    required by all steps and use `@conda` or `@pypi` to specify step-specific additions.

    Parameters
    ----------
    name : str, optional, default None
        If specified, can refer to a named environment. The environment referred to
        here will be the one used for this flow. If specified, nothing else can be
        specified in this decorator. In the name, you can use `@{}` values and
        environment variables will be used to substitute.
    pathspec : str, optional, default None
        If specified, can refer to the pathspec of an existing step. The environment
        of this referred step will be used here. If specified, nothing else can be
        specified in this decorator. In the name, you can use `@{}` values and
        environment variables will be used to substitute.
    fetch_at_exec : bool, default False
        If set to True, the environment will be fetched when the task is
        executing as opposed to at the beginning of the flow (or at deploy time if
        deploying to a scheduler). This option requires name or pathspec to be
        specified. This is useful, for example, if you want this flow to always use
        the latest named environment when it runs as opposed to the latest when it
        is deployed.
    disabled : bool, default False
        If set to True, uses the external environment.
    """

    name = "named_env_base"


class SysPackagesRequirementFlowDecorator(
    SysPackagesRequirementDecoratorMixin, PackageRequirementFlowDecorator
):
    """
    Specifies system virtual packages for this flow.

    This is an advanced usage decorator allowing you to override the __glibc and
    __cuda virtual packages that are used when resolving your environment.

    Parameters
    ----------
    packages : Dict[str, str], default {}
        System virtual packages to use for this flow. Supported keys are "__cuda" and
        "__glibc".
    disabled : bool, default False
        If set to True, uses the external environment.
    """

    name = "sys_packages_base"


class PipRequirementFlowDecorator(PypiRequirementFlowDecorator):
    """
    Specifies the Pypi packages for all steps of the flow.

    DEPRECATED: please use `@pypi_base` instead

    Parameters
    ----------
    name : str, optional, default None
        DEPRECATED -- use `@named_env(name=)` instead.
        If specified, can refer to a named environment. The environment referred to
        here will be the one used for this flow. If specified, nothing else can be
        specified in this decorator. In the name, you can use `@{}` values and
        environment variables will be used to substitute.
    pathspec : str, optional, default None
        DEPRECATED -- use `@named_env(name=)` instead.
        If specified, can refer to the pathspec of an existing step. The environment
        of this referred step will be used here. If specified, nothing else can be
        specified in this decorator. In the name, you can use `@{}` values and
        environment variables will be used to substitute.
    packages : Dict[str, str], default {}
        Packages to use for this flow. The key is the name of the package
        and the value is the version to use.
    extra_indices : List[str], default []
        Additional sources to search for
    python : str, optional, default None
        Version of Python to use, e.g. '3.7.4'. If not specified, the current python
        version will be used.
    fetch_at_exec : bool, default False
        DEPRECATED -- use `@named_env(name=)` instead.
        If set to True, the environment will be fetched when the task is
        executing as opposed to at the beginning of the flow (or at deploy time if
        deploying to a scheduler). This option requires name or pathspec to be
        specified. This is useful, for example, if you want this flow to always use
        the latest named environment when it runs as opposed to the latest when it
        is deployed.
    disabled : bool, default False
        If set to True, uses the external environment.
    """

    name = "pip_base"

    def flow_init(
        self, flow, graph, environment, flow_datastore, metadata, logger, echo, options
    ):
        echo("*DEPRECATED*: Use '@pypi_base' instead of '@%s'." % self.name)

        return super().flow_init(
            flow, graph, environment, flow_datastore, metadata, logger, echo, options
        )
